// this :
#include <BatchLab/Core/HistogramFactory.h>

#include <AIDA/IAxis.h>
#include <AIDA/ITree.h>

#include <Slash/Core/IWriter.h>
#include <Slash/Store/ITree.h>
#include <Slash/Store/IStoreManager.h>

#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Debug.h>
#include <Lib/Cast.h>
#include <Lib/Manager.h>
#include <Lib/Arguments.h>

#include <HCL/Histogram.h>

#include <BatchLab/Core/Histogram.h>
#include <BatchLab/Core/FindAllocator.h>

#define CHECK_ALLOCATOR(aCMT) \
  if(!fAllocator) {\
    Lib::Out out(fPrinter);\
    out << aCMT << " :"\
        << " null allocator." \
        << Lib::endl;\
    return 0;\
  }
  
#define THREAD_SAFETY(aCMT) \
 {Lib::Arguments args(aOptions,";",true);\
  bool thread_safe;\
  if(args.find("thread_safe",thread_safe)) {\
    BaseObject* bobject = Lib_Cast(*object,BatchLab::BaseObject);\
    if(!bobject) {\
      Lib::Out out(fPrinter);\
      out << aCMT << " :"\
          << " Slash::Store::IObject not a BatchLab::BaseObject."\
          << Lib::endl;\
      delete object;\
      return 0;\
    }\
    if(!bobject->setThreadSafe(thread_safe)){\
      Lib::Out out(fPrinter);\
      out << aCMT << " :"\
          << " unable to set thread safety."\
          << Lib::endl;\
      delete object;\
      return 0;\
    }\
  }}

//////////////////////////////////////////////////////////////////////////////
BatchLab::HistogramFactory::HistogramFactory(
 Slash::Core::ISession& aSession
,Slash::Store::ITree& aTree
)
:BaseFactory(aSession.printer(),aTree)
,fSession(aSession)
,fAllocator(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::HistogramFactory");
  fAllocator = BatchLab::histogramAllocator(fSession);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::HistogramFactory::~HistogramFactory(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::HistogramFactory");
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::createHistogram1D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  Slash::Store::IObject* object = 0;
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Slash::Store::IStoreManager* storeManager = 
      Lib_findManager(fSession,"Zebra_Manager",Slash::Store::IStoreManager);
    if(!storeManager) return 0;
    std::vector<Slash::Store::IStoreManager::Range> args;
    args.push_back
      (Slash::Store::IStoreManager::Range
        (aXnumber,std::pair<double,double>(aXmin,aXmax)));
    object = storeManager->createObject
               (fTree,"AIDA::IHistogram1D",path,name,aTitle,args,aOptions);
  } else {
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createHistogram1D")
    object = fAllocator->new_Histogram1D(fPrinter,name,aTitle,
                                             aXnumber,aXmin,aXmax);
  }
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createHistogram1D")
  AIDA::IHistogram1D* aobject = Lib_Cast(*object,AIDA::IHistogram1D);
  if(!aobject) {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram1D :" 
        << " Slash::Store::IObject not an AIDA::IHistogram1D." << Lib::endl;
    delete object;
    return 0;
  }
  if(!manage(aOptions)) return aobject;
  if(!manageObject(object,path)) {
    delete object;
    return 0;
  }
  return aobject;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::createHistogram1D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createHistogram1D(aPath,name,aXnumber,aXmin,aXmax,"");
}

//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::createHistogram1D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdges
,const std::string& aOptions
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aEdges.size()<=1) {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram1D :" 
        << " at least two edges expected" 
        << " for variable bins histogram booking." 
        << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  Slash::Store::IObject* object = 0;
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Slash::Store::IStoreManager* storeManager = 
      Lib_findManager(fSession,"Zebra_Manager",Slash::Store::IStoreManager);
    if(!storeManager) return 0;

    std::vector<Slash::Store::IStoreManager::Range> args;
    unsigned int edgen = aEdges.size();
    for(unsigned int index=0;index<edgen;index++) {
      args.push_back
        (Slash::Store::IStoreManager::Range
           (0,std::pair<double,double>(aEdges[index],0)));
    }

    object = storeManager->createObject
               (fTree,"AIDA::IHistogram1D",path,name,aTitle,args,aOptions);
  } else {
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createHistogram1D")
    object = fAllocator->new_Histogram1D(fPrinter,name,aTitle,aEdges);
  }
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createHistogram1D")
  AIDA::IHistogram1D* aobject = Lib_Cast(*object,AIDA::IHistogram1D);
  if(!aobject) {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram1D :" 
        << " Slash::Store::IObject not an AIDA::IHistogram1D." << Lib::endl;
    delete object;
    return 0;
  }
  if(!manage(aOptions)) return aobject;
  if(!manageObject(object,path)) {
    delete object;
    return 0;
  }
  return aobject;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::createHistogram2D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  Slash::Store::IObject* object = 0;
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Slash::Store::IStoreManager* storeManager = 
      Lib_findManager(fSession,"Zebra_Manager",Slash::Store::IStoreManager);
    if(!storeManager) return 0;
    std::vector<Slash::Store::IStoreManager::Range> args;
    args.push_back
      (Slash::Store::IStoreManager::Range
         (aXnumber,std::pair<double,double>(aXmin,aXmax)));
    args.push_back
      (Slash::Store::IStoreManager::Range
         (aYnumber,std::pair<double,double>(aYmin,aYmax)));
    object = storeManager->createObject
               (fTree,"AIDA::IHistogram2D",path,name,aTitle,args,aOptions);
  } else {
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createHistogram2D")
    object = fAllocator->new_Histogram2D(fPrinter,name,aTitle,
                                         aXnumber,aXmin,aXmax,
                                         aYnumber,aYmin,aYmax);
  }
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createHistogram2D")
  AIDA::IHistogram2D* aobject = Lib_Cast(*object,AIDA::IHistogram2D);
  if(!aobject) {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram2D :" 
        << " Slash::Store::IObject not an AIDA::IHistogram2D." << Lib::endl;
    delete object;
    return 0;
  }
  if(!manage(aOptions)) return aobject;
  if(!manageObject(object,path)) {
    delete object;
    return 0;
  }
  return aobject;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::createHistogram2D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createHistogram2D(aPath,name,
                           aXnumber,aXmin,aXmax,aYnumber,aYmin,aYmax,"");
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::createHistogram2D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdgesX
,const std::vector<double>& aEdgesY
,const std::string& aOptions
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram2D :" 
        << " variable bins not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createHistogram2D")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,aTitle,
                                aEdgesX,aEdgesY);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createHistogram2D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::createHistogram3D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
,int aZnumber
,double aZmin
,double aZmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram3D :" 
        << " histogram 3D not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createHistogram3D")
  BatchLab::Histogram3D* object = 
    fAllocator->new_Histogram3D(fPrinter,name,aTitle,
                                aXnumber,aXmin,aXmax,
                                aYnumber,aYmin,aYmax,
                                aZnumber,aZmin,aZmax);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createHistogram3D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::createHistogram3D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
,int aZnumber
,double aZmin
,double aZmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createHistogram3D(aPath,name,
                           aXnumber,aXmin,aXmax, 
                           aYnumber,aYmin,aYmax,
                           aZnumber,aZmin,aZmax,"");
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::createHistogram3D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdgesX
,const std::vector<double>& aEdgesY
,const std::vector<double>& aEdgesZ
,const std::string& aOptions
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createHistogram3D :" 
        << " variable bins not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createHistogram3D")
  BatchLab::Histogram3D* object = 
    fAllocator->new_Histogram3D(fPrinter,name,aTitle,
                                aEdgesX,aEdgesY,aEdgesZ);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createHistogram3D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////I
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::add( 
 const std::string& aPath
,const AIDA::IHistogram1D& aA
,const AIDA::IHistogram1D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram1D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram1D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram1D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram1D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.add(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::add")
  BatchLab::Histogram1D* object = 
    fAllocator->new_Histogram1D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::subtract( 
 const std::string& aPath
,const AIDA::IHistogram1D& aA
,const AIDA::IHistogram1D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram1D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram1D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram1D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram1D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.subtract(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::subtract")
  BatchLab::Histogram1D* object = 
    fAllocator->new_Histogram1D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::multiply( 
 const std::string& aPath
,const AIDA::IHistogram1D& aA
,const AIDA::IHistogram1D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram1D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram1D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram1D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram1D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.multiply(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::multiply")
  BatchLab::Histogram1D* object = 
    fAllocator->new_Histogram1D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::divide( 
 const std::string& aPath
,const AIDA::IHistogram1D& aA
,const AIDA::IHistogram1D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram1D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram1D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram1D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram1D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.divide(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::divide")
  BatchLab::Histogram1D* object = 
    fAllocator->new_Histogram1D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::add( 
 const std::string& aPath
,const AIDA::IHistogram2D& aA
,const AIDA::IHistogram2D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram2D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram2D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram2D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram2D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.add(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::add")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::subtract( 
 const std::string& aPath
,const AIDA::IHistogram2D& aA
,const AIDA::IHistogram2D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram2D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram2D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram2D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram2D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.subtract(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::subtract")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::multiply( 
 const std::string& aPath
,const AIDA::IHistogram2D& aA
,const AIDA::IHistogram2D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram2D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram2D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram2D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram2D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.multiply(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::multiply")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::divide( 
 const std::string& aPath
,const AIDA::IHistogram2D& aA
,const AIDA::IHistogram2D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram2D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram2D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram2D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram2D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.divide(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::divide")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::add( 
 const std::string& aPath
,const AIDA::IHistogram3D& aA
,const AIDA::IHistogram3D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram3D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram3D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram3D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.add(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::add")
  BatchLab::Histogram3D* object = 
    fAllocator->new_Histogram3D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::subtract( 
 const std::string& aPath
,const AIDA::IHistogram3D& aA
,const AIDA::IHistogram3D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram3D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram3D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram3D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.subtract(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::subtract")
  BatchLab::Histogram3D* object = 
    fAllocator->new_Histogram3D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::multiply( 
 const std::string& aPath
,const AIDA::IHistogram3D& aA
,const AIDA::IHistogram3D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram3D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram3D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram3D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.multiply(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::multiply")
  BatchLab::Histogram3D* object = 
    fAllocator->new_Histogram3D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::divide( 
 const std::string& aPath
,const AIDA::IHistogram3D& aA
,const AIDA::IHistogram3D& aB
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabA = 
    Lib_ConstCast(aA,BatchLab::Histogram3D);
  if(!batchlabA) return 0;
  const HCL::Histogram& bareA = batchlabA->bareHistogram();
  const BatchLab::Histogram3D* batchlabB = 
    Lib_ConstCast(aB,BatchLab::Histogram3D);
  if(!batchlabB) return 0;
  const HCL::Histogram& bareB = batchlabB->bareHistogram();
  HCL::Histogram bareResult(bareA);
  if(!bareResult.divide(bareB)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::divide")
  BatchLab::Histogram3D* object = 
    fAllocator->new_Histogram3D(fPrinter,name,bareResult);
  if(!object) return 0;
  object->setTitle(name);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::sliceX( 
 const std::string& aPath
,const AIDA::IHistogram2D& aHisto
,int aIndexY1
,int aIndexY2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram2D* batchlabHisto = 
    Lib_ConstCast(aHisto,BatchLab::Histogram2D);
  if(!batchlabHisto) return 0;
  const HCL::Histogram2D& bareHisto = batchlabHisto->hclHistogram();
  HCL::Base1D* bareResult = bareHisto.sliceX(aIndexY1,aIndexY2);
  if(!bareResult) return 0;
  bareResult->setTitle(name);
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::sliceX")
  BatchLab::Histogram1D* object = 
    fAllocator->new_Histogram1D(fPrinter,name,*bareResult);
  delete bareResult;
  if(!object) return 0;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::sliceY( 
 const std::string& aPath
,const AIDA::IHistogram2D& aHisto
,int aIndexX1
,int aIndexX2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram2D* batchlabHisto = 
    Lib_ConstCast(aHisto,BatchLab::Histogram2D);
  if(!batchlabHisto) return 0;
  const HCL::Histogram2D& bareHisto = batchlabHisto->hclHistogram();
  HCL::Base1D* bareResult = bareHisto.sliceY(aIndexX1,aIndexX2);
  if(!bareResult) return 0;
  bareResult->setTitle(name);
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::sliceY")
  BatchLab::Histogram1D* object = 
    fAllocator->new_Histogram1D(fPrinter,name,*bareResult);
  delete bareResult;
  if(!object) return 0;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::sliceX( 
 const std::string& aPath
,const AIDA::IHistogram2D& aHisto
,int aIndexY
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceX(aPath,aHisto,aIndexY,aIndexY);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::sliceY( 
 const std::string& aPath
,const AIDA::IHistogram2D& aHisto
,int aIndexX
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceY(aPath,aHisto,aIndexX,aIndexX);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::projectionX( 
 const std::string& aPath
,const AIDA::IHistogram2D& aHisto
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceX(aPath,aHisto,
                AIDA::IAxis::UNDERFLOW_BIN,
                AIDA::IAxis::OVERFLOW_BIN);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::projectionY(
 const std::string& aPath
,const AIDA::IHistogram2D& aHisto
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceY(aPath,aHisto,
                AIDA::IAxis::UNDERFLOW_BIN,
                AIDA::IAxis::OVERFLOW_BIN);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::sliceXY( 
 const std::string& aPath
,const AIDA::IHistogram3D& aHisto
,int aIndexZ1
,int aIndexZ2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabHisto = 
    Lib_ConstCast(aHisto,BatchLab::Histogram3D);
  if(!batchlabHisto) return 0;
  const HCL::Histogram3D& bareHisto = batchlabHisto->hclHistogram();
  HCL::Base2D* bareResult = bareHisto.sliceXY(aIndexZ1,aIndexZ2);
  if(!bareResult) return 0;
  bareResult->setTitle(name);
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::sliceXY")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,*bareResult);
  delete bareResult;
  if(!object) return 0;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::sliceXZ( 
 const std::string& aPath
,const AIDA::IHistogram3D& aHisto
,int aIndexY1
,int aIndexY2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabHisto = 
    Lib_ConstCast(aHisto,BatchLab::Histogram3D);
  if(!batchlabHisto) return 0;
  const HCL::Histogram3D& bareHisto = batchlabHisto->hclHistogram();
  HCL::Base2D* bareResult = bareHisto.sliceXZ(aIndexY1,aIndexY2);
  if(!bareResult) return 0;
  bareResult->setTitle(name);
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::sliceXZ")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,*bareResult);
  delete bareResult;
  if(!object) return 0;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::sliceYZ( 
 const std::string& aPath
,const AIDA::IHistogram3D& aHisto
,int aIndexX1
,int aIndexX2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  const BatchLab::Histogram3D* batchlabHisto = 
    Lib_ConstCast(aHisto,BatchLab::Histogram3D);
  if(!batchlabHisto) return 0;
  const HCL::Histogram3D& bareHisto = batchlabHisto->hclHistogram();
  HCL::Base2D* bareResult = bareHisto.sliceYZ(aIndexX1,aIndexX2);
  if(!bareResult) return 0;
  bareResult->setTitle(name);
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::sliceYZ")
  BatchLab::Histogram2D* object = 
    fAllocator->new_Histogram2D(fPrinter,name,*bareResult);
  delete bareResult;
  if(!object) return 0;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::projectionXY( 
 const std::string& aPath
,const AIDA::IHistogram3D& aHisto
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceXY(aPath,aHisto,
                 AIDA::IAxis::UNDERFLOW_BIN,
                 AIDA::IAxis::OVERFLOW_BIN);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::projectionXZ(
 const std::string& aPath
,const AIDA::IHistogram3D& aHisto
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceXZ(aPath,aHisto,
                 AIDA::IAxis::UNDERFLOW_BIN,
                 AIDA::IAxis::OVERFLOW_BIN);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::projectionYZ( 
 const std::string& aPath
,const AIDA::IHistogram3D& aHisto
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return sliceYZ(aPath,aHisto,
                 AIDA::IAxis::UNDERFLOW_BIN,
                 AIDA::IAxis::OVERFLOW_BIN);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include <BatchLab/Core/Cloud.h>
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud1D* BatchLab::HistogramFactory::createCloud1D(
 const std::string& aPath
,const std::string& aTitle
,int aLimit
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createCloud1D :" 
        << " cloud not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCloud1D")
  BatchLab::Cloud1D* object = 
    fAllocator->new_Cloud1D(fPrinter,name,aTitle,aLimit);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createCloud1D")
  if(!manage(aOptions)) return object;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud1D* BatchLab::HistogramFactory::createCloud1D(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createCloud1D(aPath,name,-1,"");
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud2D* BatchLab::HistogramFactory::createCloud2D(
 const std::string& aPath
,const std::string& aTitle
,int aLimit
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createCloud2D :" 
        << " cloud not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCloud2D")
  BatchLab::Cloud2D* object = 
    fAllocator->new_Cloud2D(fPrinter,name,aTitle,aLimit);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createCloud2D")
  if(!manage(aOptions)) return object;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud2D* BatchLab::HistogramFactory::createCloud2D(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createCloud2D(aPath,name,-1,"");
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud3D* BatchLab::HistogramFactory::createCloud3D(
 const std::string& aPath
,const std::string& aTitle
,int aLimit 
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createCloud3D :" 
        << " cloud not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCloud3D")
  BatchLab::Cloud3D* object = 
    fAllocator->new_Cloud3D(fPrinter,name,aTitle,aLimit);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createCloud3D")
  if(!manage(aOptions)) return object;
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud3D* BatchLab::HistogramFactory::createCloud3D(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createCloud3D(aPath,name,-1,"");
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram1D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::IHistogram1D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Histogram1D* from  = Lib_ConstCast(aFrom,BatchLab::Histogram1D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Histogram1D* object = fAllocator->new_Histogram1D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram2D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::IHistogram2D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Histogram2D* from = Lib_ConstCast(aFrom,BatchLab::Histogram2D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Histogram2D* object = fAllocator->new_Histogram2D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogram3D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::IHistogram3D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Histogram3D* from = Lib_ConstCast(aFrom,BatchLab::Histogram3D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Histogram3D* object = fAllocator->new_Histogram3D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::HistogramFactory::destroy(
 AIDA::IBaseHistogram* aHistogram
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aHistogram) return false;
  Slash::Store::IObject* bo = Lib_Cast(*aHistogram,Slash::Store::IObject);
  if(!bo) return false;
  if(!fTree.isObjectValid(bo)) return false;
  delete bo; //It remove the object from the folder (See ~BaseObject)
  fTree.emitUpdate();
  return true;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud1D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::ICloud1D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Cloud1D* from  = Lib_ConstCast(aFrom,BatchLab::Cloud1D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Cloud1D* object = fAllocator->new_Cloud1D(*from);
    if(!object) return 0;      
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud2D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::ICloud2D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Cloud2D* from  = Lib_ConstCast(aFrom,BatchLab::Cloud2D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Cloud2D* object = fAllocator->new_Cloud2D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ICloud3D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::ICloud3D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Cloud3D* from  = Lib_ConstCast(aFrom,BatchLab::Cloud3D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Cloud3D* object = fAllocator->new_Cloud3D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createProfile1D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile1D :" 
        << " profile 1D without passing YMIN, YMAX"
        << " not supported by BatchLab::Zebra." 
        << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile1D")
  BatchLab::Profile1D* object = 
    fAllocator->new_Profile1D(fPrinter,name,aTitle,
                              aXnumber,aXmin,aXmax);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile1D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createProfile1D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,double aVmin
,double aVmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  Slash::Store::IObject* object = 0;
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Slash::Store::IStoreManager* storeManager = 
      Lib_findManager(fSession,"Zebra_Manager",Slash::Store::IStoreManager);
    if(!storeManager) return 0;
    std::vector<Slash::Store::IStoreManager::Range> args;
    args.push_back
      (Slash::Store::IStoreManager::Range
         (aXnumber,std::pair<double,double>(aXmin,aXmax)));
    args.push_back
      (Slash::Store::IStoreManager::Range
         (0,std::pair<double,double>(aVmin,aVmax)));
    object = storeManager->createObject
               (fTree,"AIDA::IProfile1D",path,name,aTitle,args,aOptions);
  } else {
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile1D")
    object = fAllocator->new_Profile1D(fPrinter,name,aTitle,
                                       aXnumber,aXmin,aXmax,
                                       aVmin,aVmax);
  }
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile1D")
  AIDA::IProfile1D* aobject = Lib_Cast(*object,AIDA::IProfile1D);
  if(!aobject) {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile1D :" 
        << " Slash::Store::IObject not an AIDA::IProfile1D." << Lib::endl;
    delete object;
    return 0;
  }
  if(!manage(aOptions)) return aobject;
  if(!manageObject(object,path)) {
    delete object;
    return 0;
  }
  return aobject;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createProfile1D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdges
,const std::string& aOptions
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile1D :" 
        << " profile 1D with variable size bins"
        << " not supported by BatchLab::Zebra." 
        << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile1D")
  BatchLab::Profile1D* object = 
    fAllocator->new_Profile1D(fPrinter,name,aTitle,aEdges);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile1D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createProfile1D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdges
,double aVmin
,double aVmax
,const std::string& aOptions
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile1D :" 
        << " profile 1D with variable size bins"
        << " not supported by BatchLab::Zebra." 
        << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile1D")
  BatchLab::Profile1D* object = 
    fAllocator->new_Profile1D(fPrinter,name,aTitle,
                              aEdges,aVmin,aVmax);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile1D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createProfile1D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createProfile1D(aPath,name,aXnumber,aXmin,aXmax,"");
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createProfile1D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
,double aVmin
,double aVmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createProfile1D(aPath,name,aXnumber,aXmin,aXmax,aVmin,aVmax,"");
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createProfile2D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile2D :" 
        << " profile 2D not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile2D")
  BatchLab::Profile2D* object = 
    fAllocator->new_Profile2D(fPrinter,name,aTitle,
                              aXnumber,aXmin,aXmax,
                              aYnumber,aYmin,aYmax);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile2D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createProfile2D(
 const std::string& aPath
,const std::string& aTitle
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
,double aVmin
,double aVmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile2D :" 
        << " profile 2D not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile2D")
  BatchLab::Profile2D* object = 
    fAllocator->new_Profile2D(fPrinter,name,aTitle,
                              aXnumber,aXmin,aXmax,
                              aYnumber,aYmin,aYmax,
                              aVmin,aVmax);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile2D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createProfile2D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createProfile2D(aPath,name,
                         aXnumber,aXmin,aXmax,aYnumber,aYmin,aYmax,"");
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createProfile2D(
 const std::string& aPath
,int aXnumber
,double aXmin
,double aXmax
,int aYnumber
,double aYmin
,double aYmax
,double aVmin
,double aVmax
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  return createProfile2D(aPath,name,
                    aXnumber,aXmin,aXmax,aYnumber,aYmin,aYmax,aVmin,aVmax,"");
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createProfile2D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdgesX
,const std::vector<double>& aEdgesY
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile2D :" 
        << " profile 2D not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile2D")
  BatchLab::Profile2D* object = 
    fAllocator->new_Profile2D(fPrinter,name,aTitle,aEdgesX,aEdgesY);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile2D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createProfile2D(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<double>& aEdgesX
,const std::vector<double>& aEdgesY
,double aVmin
,double aVmax
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    Lib::Out out(fPrinter);
    out << "BatchLab::HistogramFactory::createProfile2D :" 
        << " profile 2D not supported by HBOOK." << Lib::endl;
    return 0;
  }
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  CHECK_ALLOCATOR("BatchLab::HistogramFactory::createProfile2D")
  BatchLab::Profile2D* object = 
    fAllocator->new_Profile2D(fPrinter,name,aTitle,
                              aEdgesX,aEdgesY,aVmin,aVmax);
  if(!object) return 0;
  THREAD_SAFETY("BatchLab::HistogramFactory::createProfile2D")
  if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile1D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::IProfile1D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Profile1D* from  = Lib_ConstCast(aFrom,BatchLab::Profile1D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Profile1D* object = fAllocator->new_Profile1D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IProfile2D* BatchLab::HistogramFactory::createCopy(
 const std::string& aPath
,const AIDA::IProfile2D& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(const Profile2D* from  = Lib_ConstCast(aFrom,BatchLab::Profile2D)) {
    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
    CHECK_ALLOCATOR("BatchLab::HistogramFactory::createCopy")
    BatchLab::Profile2D* object = fAllocator->new_Profile2D(*from);
    if(!object) return 0;
    object->setName(name);
    //if(!manage(aOptions)) return object;
    if(manageObject(object,path)) return object;
    delete object;
    return 0;
  } else {
    return 0;//FIXME
  }
}
