/**************************************************************************
 *
 *  This file is part of the Coin SoWin GUI binding library.
 *  Copyright (C) 2000 by Systems in Motion.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  version 2.1 as published by the Free Software Foundation.  See the
 *  file LICENSE.LGPL at the root directory of the distribution for
 *  more details.
 *
 *  If you want to use Coin SoWin for applications not compatible with the
 *  LGPL, please contact SIM to aquire a Professional Edition License.
 *
 *  Systems in Motion, Prof Brochs gate 6, N-7030 Trondheim, NORWAY
 *  http://www.sim.no/ support@sim.no Voice: +47 22114160 Fax: +47 22207097
 *
 **************************************************************************/

static const char rcsid[] =
  "$Id: WinNativePopupMenu.cpp,v 1.2 2002/02/21 13:57:13 barrand Exp $";

#include <windows.h>

#include <Inventor/SoLists.h>
#include <Inventor/errors/SoDebugError.h>

#include <sowindefs.h>
#include <Inventor/Win/widgets/WinNativePopupMenu.h>
#include <Inventor/Win/Win32API.h>

/*!
  \class WinNativePopupMenu Inventor/Win/widgets/WinNativePopupMenu.h
  \brief The WinNativePopupMenu class implements a common interface for popup
  menu management for all the Coin GUI toolkit libraries.
*/

// *************************************************************************

struct MenuRecord {
  int menuid;
  char * name;
  char * title;
  HMENU menu;
  HMENU parent;
}; // struct MenuRecord

struct ItemRecord {
  int itemid;
  int flags;
  char * name;
  char * title;
  HMENU parent;
}; // struct ItemRecord

#define ITEM_TOGGLE       0x0001
#define ITEM_MARKED       0x0002
#define ITEM_SEPARATOR    0x0004
#define ITEM_ENABLED      0x0008

// *************************************************************************

WinNativePopupMenu::WinNativePopupMenu(void)
{
  this->menus = new SbPList;
  this->items = new SbPList;

  this->notify = FALSE;
  this->selectedItem = -1;
} // WinNativePopupMenu()

WinNativePopupMenu::~WinNativePopupMenu(void)
{
  const int numMenus = this->menus->getLength();
  HMENU popup = NULL;
  int i;
  for (i = 0; i < numMenus; i++) {
    MenuRecord * rec = (MenuRecord *) (* this->menus)[i];
    if (rec->menuid == 0) popup = rec->menu;
    delete [] rec->name;
    delete [] rec->title;
    if (rec->parent == NULL) delete rec->menu; // menu not attached
    delete rec;
  }

  const int numItems = this->items->getLength();
  for (i = 0; i < numItems; i++) {
    ItemRecord * rec = (ItemRecord *) (* this->items)[i];
    delete [] rec->name;
    delete [] rec->title;
    delete rec;
  }

  // delete root popup menu
  delete popup;
} // ~WinNativePopupMenu()

// *************************************************************************

int
WinNativePopupMenu::newMenu(const char * name, int menuid)
{
  int id = menuid;
  if (id == -1) {
    id = 1;
    while (this->getMenuRecord(id) != NULL) id++;
  } else {
    assert(this->getMenuRecord(id) == NULL &&
            "requested menuid already taken");
  }
  // id contains ok ID
  MenuRecord * rec = createMenuRecord(name);
  rec->menuid = id;
  this->menus->append((void *) rec);
  return id;
} // newMenu()

int
WinNativePopupMenu::getMenu(const char * name)
{
  const int numMenus = this->menus->getLength();
  int i;
  for (i = 0; i < numMenus; i++) {
    if (strcmp(((MenuRecord *) (* this->menus)[i])->name, name) == 0) {
      return ((MenuRecord *) (* this->menus)[i])->menuid;
    }
  }
  return -1;
} // getMenu()

void
WinNativePopupMenu::setMenuTitle(int menuid, const char * title)
{
  MenuRecord * rec = this->getMenuRecord(menuid);
  assert(rec != NULL && "no such menu");
  delete [] rec->title;
  rec->title = strcpy(new char [strlen(title)+1], title);

  if (rec->parent)
    Win32::ModifyMenu(rec->parent, rec->menuid, MF_BYPOSITION | MF_STRING, rec->menuid, rec->title);

} // setMenuTitle()

const char *
WinNativePopupMenu::getMenuTitle(int menuid)
{
  MenuRecord * rec = this->getMenuRecord(menuid);
  assert(rec != NULL && "no such menu");
  return rec->title;
} // getMenuTitle()

// *************************************************************************

int
WinNativePopupMenu::newMenuItem(const char * name, int itemid)
{
  int id = itemid;
  if (id == -1) {
    id = 1;
    while (this->getItemRecord(itemid) != NULL) id++;
  } else {
    if (this->getItemRecord(itemid) != NULL) {
#if SOWIN_DEBUG
      SoDebugError::postInfo("WinNativePopupMenu::newMenuItem",
                              "requested itemid already taken");
#endif // SOWIN_DEBUG
      return -1;
    }
  }
  ItemRecord * rec = createItemRecord(name);
  rec->itemid = id;
  this->items->append(rec);
  return id;
} // newMenuItem()

int
WinNativePopupMenu::getMenuItem(const char * name)
{
  const int numItems = this->items->getLength();
  int i;
  for (i = 0; i < numItems; i++)
    if (strcmp(((ItemRecord *) (* this->items)[i])->name, name) == 0)
      return ((ItemRecord *) (* this->items)[i])->itemid;
  return -1;
} // getMenuItem()

void
WinNativePopupMenu::setMenuItemTitle(int itemid, const char * title)
{
  ItemRecord * rec = this->getItemRecord(itemid);
  assert(rec != NULL && "no such menu");
  delete [] rec->title;
  rec->title = strcpy(new char [strlen(title)+1], title);

  if (rec->parent)
    Win32::ModifyMenu(rec->parent, rec->itemid, MF_BYCOMMAND | MF_STRING, rec->itemid, rec->title);
} // setMenuItemTitle()

const char *
WinNativePopupMenu::getMenuItemTitle(int itemid)
{
  ItemRecord * rec = this->getItemRecord(itemid);
  assert(rec != NULL && "no such menu");
  return rec->title;
} // getMenuItemTitle()

void
WinNativePopupMenu::setMenuItemEnabled(int itemid, SbBool enabled)
{
  ItemRecord * rec = this->getItemRecord(itemid);

  assert(rec && "could not find item record");
  assert(rec->parent && "a menuitem must have a parent to be enabled/disabled");
  assert(IsMenu(rec->parent));
  
  MENUITEMINFO info;
  // (no need to ZeroMemory() structure)

  info.cbSize = sizeof(MENUITEMINFO);
  info.fMask = MIIM_STATE;
  
  if (enabled) {
    rec->flags |= ITEM_ENABLED;
    info.fState = MFS_ENABLED;
  }
  else {
    rec->flags &= ~ITEM_ENABLED;    
    info.fState = MFS_DISABLED | MFS_GRAYED;
  }
  
  Win32::SetMenuItemInfo(rec->parent, rec->itemid, FALSE, & info);
} // setMenuItemEnabled()

SbBool
WinNativePopupMenu::getMenuItemEnabled(int itemid)
{
  ItemRecord * rec = this->getItemRecord(itemid);
  
  assert(rec && "could not find item record");
  assert(IsMenu(rec->parent));

  //MENUITEMINFO  menuiteminfo;
  //memset((void *) & menuiteminfo, 0, sizeof(menuiteminfo));
  //Win32::GetMenuItemInfo(rec->parent, rec->itemid, TRUE, & menuiteminfo);
  //return (menuiteminfo.fState & MFS_ENABLED) ? TRUE : FALSE;
 
  return (rec->flags & ITEM_ENABLED ? TRUE : FALSE);
} // getMenuItemEnabled()

void
WinNativePopupMenu::_setMenuItemMarked(int itemid, SbBool marked)
{
  ItemRecord * rec = this->getItemRecord(itemid);

  assert(rec != NULL && "no such menu");
  assert(IsMenu(rec->parent));

  rec->flags |= ITEM_TOGGLE;

  MENUITEMINFO info;
  
  info.cbSize = sizeof(MENUITEMINFO);
  info.fMask = MIIM_STATE;
  
  if (marked) {
    rec->flags |= ITEM_MARKED;
    info.fState = MFS_CHECKED;
  }
  else {
    rec->flags &= ~ITEM_MARKED;    
    info.fState = MFS_UNCHECKED;
  }
  
  Win32::SetMenuItemInfo(rec->parent, rec->itemid, FALSE, & info);

} // setMenuItemMarked()

SbBool
WinNativePopupMenu::getMenuItemMarked(int itemid)
{
  ItemRecord * rec = this->getItemRecord(itemid);
  assert(rec != NULL && "no such menu");
  assert(rec->parent != NULL);
  
  MENUITEMINFO info;

  info.cbSize = sizeof(MENUITEMINFO);
  info.fMask = MIIM_STATE;
  
  Win32::GetMenuItemInfo(rec->parent, rec->itemid, FALSE, & info);

  return (info.fState & MFS_CHECKED ? TRUE : FALSE);
} // getMenuItemMarked()

// *************************************************************************

void
WinNativePopupMenu::addMenu(int menuid, int submenuid, int pos)
{
  MenuRecord * super = this->getMenuRecord(menuid);
  MenuRecord * sub = this->getMenuRecord(submenuid);
  assert(super != NULL && sub != NULL && "no such menu");

  MENUITEMINFO menuiteminfo;
  memset((void *) & menuiteminfo, 0, sizeof(menuiteminfo));
  menuiteminfo.cbSize = sizeof(menuiteminfo);
  menuiteminfo.fMask = MIIM_SUBMENU | MIIM_TYPE;
  menuiteminfo.fType = MFT_STRING;
  menuiteminfo.hSubMenu = sub->menu;
  menuiteminfo.dwTypeData = sub->title;
  menuiteminfo.cch = strlen(sub->title);

  if (pos == -1)
    Win32::InsertMenuItem(super->menu, sub->menuid, FALSE, & menuiteminfo);
  else
    Win32::InsertMenuItem(super->menu, pos, TRUE, & menuiteminfo);
  sub->parent = super->menu;
} // addMenu()

void
WinNativePopupMenu::addMenuItem(int menuid, int itemid, int pos)
{
  MenuRecord * menu = this->getMenuRecord(menuid);
  ItemRecord * item = this->getItemRecord(itemid);
  assert(menu != NULL && item != NULL && "no such menu");
  
  Win32::InsertMenu(menu->menu, pos, MF_BYPOSITION | MF_STRING, item->itemid, item->title);

  item->parent = menu->menu;
  if (item->flags & ITEM_MARKED)
    Win32::CheckMenuItem(item->parent, item->itemid, MF_BYCOMMAND | MF_CHECKED);
} // addMenuItem()

void
WinNativePopupMenu::addSeparator(int menuid, int pos)
{
  MenuRecord * menu = this->getMenuRecord(menuid);
  assert(menu != NULL && "no such menu");
  ItemRecord * rec = createItemRecord("separator");
 
  Win32::InsertMenu(menu->menu, pos, MF_BYPOSITION | MF_SEPARATOR, pos, NULL);
  rec->flags |= ITEM_SEPARATOR;
  this->items->append(rec);
} // addSeparator()

void
WinNativePopupMenu::removeMenu(int menuid)
{
  MenuRecord * rec = this->getMenuRecord(menuid);
  assert(rec != NULL && "no such menu");

  // FIXME: just assumes root-menu has id==0. Bad. 20010810 mortene.
  if (rec->menuid == 0) {
#if SOWIN_DEBUG
    SoDebugError::postInfo("WinNativePopupMenu::removeMenu", "can't remove root");
#endif // SOWIN_DEBUG
    return;
  }
  if (rec->parent == NULL) {
#if SOWIN_DEBUG
    SoDebugError::postInfo("WinNativePopupMenu::removeMenu", "menu not attached");
#endif // SOWIN_DEBUG
    return;
  }
  Win32::RemoveMenu(rec->menu, rec->menuid, MF_BYCOMMAND);
  rec->parent = NULL;
} // removeMenu()

void
WinNativePopupMenu::removeMenuItem(int itemid)
{
  ItemRecord * rec = this->getItemRecord(itemid);
  assert(rec != NULL && "no such menu");
  if (rec->parent == NULL) {
#if SOWIN_DEBUG
    SoDebugError::postInfo("WinNativePopupMenu::removeMenuItem", "item not attached");
#endif // SOWIN_DEBUG
    return;
  }
  Win32::RemoveMenu(rec->parent, rec->itemid, MF_BYCOMMAND);
  rec->parent = NULL;
} // removeMenuItem()

// *************************************************************************

void
WinNativePopupMenu::popUp(HWND inside, int x, int y)
{
  POINT point;
  point.x = x;
  point.y = y;
  Win32::ClientToScreen(inside, &point);

  MenuRecord * menurec = this->getMenuRecord(0);
  this->selectedItem = TrackPopupMenu(menurec->menu,
                                      TPM_LEFTALIGN |
                                      TPM_TOPALIGN |
                                      TPM_RIGHTBUTTON |
                                      TPM_RETURNCMD |
                                      (this->notify ? 0 : TPM_NONOTIFY),
                                      point.x,
                                      point.y,
                                      0,
                                      inside,
                                      NULL);
  
  if (this->selectedItem == 0)
    return;
  
  this->invokeMenuSelection(this->selectedItem);

  ItemRecord * itemrec = this->getItemRecord(this->selectedItem);
  assert(itemrec != NULL);
} // popUp()

int
WinNativePopupMenu::getSelectedItem(void)
{
 return (this->selectedItem);
} // getSelectedItem()

void
WinNativePopupMenu::setNotify(SbBool enable)
{
 this->notify = enable;
} // setNotify()

// *************************************************************************

MenuRecord *
WinNativePopupMenu::getMenuRecord(int menuid)
{
  const int numMenus = this->menus->getLength();
  int i;
  for (i = 0; i < numMenus; i++)
    if (((MenuRecord *) (* this->menus)[i])->menuid == menuid)
      return (MenuRecord *) (* this->menus)[i];
  return (MenuRecord *) NULL;
} // getMenuRecord()

ItemRecord *
WinNativePopupMenu::getItemRecord(int itemid)
{
  const int numItems = this->items->getLength();
  int i;
 
  for (i = 0; i < numItems; i++)
    if (((ItemRecord *) (* this->items)[i])->itemid == itemid)
      return (ItemRecord *) (* this->items)[i];

  return (ItemRecord *) NULL;
} // getItemRecord()

// *************************************************************************

MenuRecord *
WinNativePopupMenu::createMenuRecord(const char * name)
{
  MenuRecord * rec = new MenuRecord;
  rec->menuid = -1;
  rec->name = strcpy(new char [strlen(name) + 1], name);
  rec->title = strcpy(new char [strlen(name) + 1], name);
  rec->menu = CreatePopupMenu();
  rec->parent = NULL;
  return rec;
} // create()

ItemRecord *
WinNativePopupMenu::createItemRecord(const char * name)
{
  ItemRecord * rec = new ItemRecord;
  rec->itemid = -1;
  rec->flags = 0;
  rec->name = strcpy(new char [strlen(name) + 1], name);
  rec->title = strcpy(new char [strlen(name) + 1], name);
  rec->parent = NULL;
  return rec;
} // create()

// *************************************************************************

void
WinNativePopupMenu::itemActivation(int itemid)
{
  inherited::invokeMenuSelection(itemid);
} // menuSelection()

// *************************************************************************
