// There is no explicit dependancy to G4 libs
// due to the fact that we can't have DLLs for them
// on Windows and Darwin. We pass through the
// IGeant4Manager interface and the Lib accessor manager 
// to access Geant4 from the GUI callbacks.

#include <Slash/UI/IUI.h>
#include <Slash/Tools/smanip.h>
#include <Slash/Tools/args.h>
#include <Slash/Tools/Manager.h>

// G4Lab :
#include <G4Lab/Interfaces/IGeant4Manager.h>

inline IGeant4Manager* find_Geant4Manager(Slash::Core::ISession& aSession) {
  IGeant4Manager* g4Manager = 
    Slash_findManager(aSession,"Geant4Manager",IGeant4Manager);
  if(!g4Manager) {
    std::ostream& out = aSession.cout();
    out << "Geant4Manager not found." << std::endl;
    return 0;
  }
  return g4Manager;
}

extern "C" {

//////////////////////////////////////////////////////////////////////////////
void G4Lab_declare_accessors(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  //Declare G4 accessors (PV, Trajectory, HitsCollections, etc...).
  g4Manager->initialize(); 
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab_exit(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  if(g4Manager->isRunning()) {
    g4Manager->abortRun();
    aUI.exit();
    Slash::UI::IUI::Options params;
    //FIXME : background needed ?
    params.push_back(Slash::UI::IUI::Option("background","yes"));
    aUI.executeScript("DLD","OnX ui_exit",params);
  } else {
    aUI.exit();
  }
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab_next_event(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  if(!g4Manager->isRunning()) {
    aUI.warn("No run started.");
  } else {
    aUI.exit();
  }
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab_set_script(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] [event_begin,event_end,run_begin,run_end]
// args[1] interpreter
// args[2,n] script
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!Slash::check_min(aArgs,3,last,aUI.session().cout())) return;
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  const std::string& arg0 = aArgs[0];
  if(arg0=="event_begin") {
    g4Manager->setEventBeginScript(aArgs[1],last);
  } else if(arg0=="event_end") {
    g4Manager->setEventEndScript(aArgs[1],last);
  } else if(arg0=="run_begin") {
    g4Manager->setRunBeginScript(aArgs[1],last);
  } else if(arg0=="run_end") {
    g4Manager->setRunEndScript(aArgs[1],last);
  } else {
    std::ostream& out = aUI.session().cout();
    out << "G4Lab_set_script :" 
        << " unknown keyword " << Slash::sout(arg0)
        << std::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab_dump_particle_tables(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  g4Manager->particleTableDump();
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab_verbose_off(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  g4Manager->verboseOff();
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab_execute_script(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] G4 script.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!Slash::check_args(aArgs,1,aUI.session().cout())) return;
  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  g4Manager->executeG4Script(aArgs[0]);
}

} // extern "C"
