
// this :
#include <G4Lab/Manager.h>

// Inventor :
#include <Inventor/nodes/SoNode.h>

#ifdef WIN32
#undef pascal // Clash between windef.h and Geant4/SystemOfUnits.hh
#endif

// Geant4 :
#include <G4UImanager.hh>
#include <G4UIsession.hh>
#include <G4RunManager.hh>
#include <G4UIcommandTree.hh>
#include <G4SDManager.hh>
#include <G4DigiManager.hh>

// Slash :
#include <Slash/Core/IManager.h>
#include <Slash/Core/ISession.h>
#include <Slash/Data/IProcessor.h>
#include <Slash/Tools/Managers.h>
#include <Slash/UI/IScriptManager.h>
#include <Slash/Tools/smanip.h>

// Lib :
#include <Lib/Cast.h>

// AIDA :
#include <AIDA/IAnalysisFactory.h>

// G4Lab :
#include <G4Lab/ParticleAccessor.h>
#include <G4Lab/PhysicalVolumeAccessor.h>
#include <G4Lab/TrajectoryContainerAccessor.h>
#include <G4Lab/TrajectoryAccessor.h>
#include <G4Lab/PhysicsTableAccessor.h>
#include <G4Lab/HitsCollectionAccessor.h>
#include <G4Lab/Tree.h>
#include <G4Lab/State.h>
#include <G4Lab/SoG4Trajectories.h>
#include <G4Lab/SoG4RunManager.h>

// From G4VBasicShell (tag geant4-05-00-ref-01)
static G4String Complete(G4String);
static G4String FindMatchingPath(G4UIcommandTree*,G4String);

// Class to redirect G4 output to OnX console and do command completion :
// Avoid establishing a relationship with G4interfaces (by inheriting
// G4VBasicShell). This library may be reconstructed with various GUI drivers.

namespace G4Lab {
class UIsession : public G4UIsession {
public: //G4UIsession
  virtual G4int ReceiveG4cout(G4String aString){
    fOut << aString;
    return 0;
  }
  virtual G4int ReceiveG4cerr(G4String aString){
    fOut << aString;
    return 0;
  }
public:
  UIsession(std::ostream& aOut):fOut(aOut){}
  std::string completeCommand(const std::string& aString) {
    return Complete(aString);
  }
private:
  std::ostream& fOut;
};

}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
G4Lab::Manager::Manager(
 Slash::Core::ISession& aSession                    
,const std::string& aName
,G4RunManager* aRunManager
,bool aDeleteRunManager
)
:fSession(aSession)
,fName(aName)
,fUIsession(0)
,fState(0)
,fScriptManager(0)
,fRunManager(aRunManager)
,fDeleteRunManager(aDeleteRunManager)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4UImanager* UI = G4UImanager::GetUIpointer();
  if(UI) {
    fUIsession = new UIsession(fSession.cout());
    UI->SetSession(fUIsession);
    UI->SetCoutDestination(fUIsession);
  }

  fScriptManager = Slash::scriptManager(fSession);
  if(!fScriptManager) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::Manager :"
        << " ScriptManager not found." 
        << std::endl;
  } else {
    //FIXME : command to exec a mac file ?
    fScriptManager->addInterpreter("G4","","","mac","","",
                                   executeScript,0,completer,0); 
  }

  // The G4Lab::State will be deleted by the run manager.
  fState = new G4Lab::State(fScriptManager);
}
//////////////////////////////////////////////////////////////////////////////
G4Lab::Manager::~Manager(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4UImanager* UI = G4UImanager::GetUIpointer();
  if(UI) {
    UI->SetSession(0);
    UI->SetCoutDestination(0);
  }
  delete fUIsession;

  // We should remove G4Lab things in the AccessorManager...

  // The G4Lab::State is deleted by the run manager.

  if(fDeleteRunManager) {
    delete fRunManager;
    fRunManager = 0;
  } else {
    // The fRunManager is deleted by the user.
  }
}

#include "find_AIDA.cxx"

AIDA::IAnalysisFactory* G4Lab::Manager::findAIDA() {
  return find_AIDA(fSession);
}

//////////////////////////////////////////////////////////////////////////////
bool G4Lab::Manager::initialize(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //NOTE : this function should be reexcutable.

  // So Should be initialized once after Inventor and HEPVis :
  if(SoG4Trajectories::getClassTypeId()==SoType::badType()) {
    SoG4Trajectories::initClass();
  }
  if(SoG4RunManager::getClassTypeId()==SoType::badType()) {
    SoG4RunManager::initClass();
  }

  Slash::Data::IProcessor* da = Slash::accessorManager(fSession);
  if(!da) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::initialize :"
	<< " AccessorManager not found." << std::endl;
    return false;
  }

  if(!da->findAccessor(PhysicalVolumeAccessor::accessor_name())) {
    da->addAccessor(new PhysicalVolumeAccessor(fSession));
  }
  if(!da->findAccessor(ParticleAccessor::accessor_name())) {
    da->addAccessor(new ParticleAccessor(fSession));
  }
  if(!da->findAccessor(PhysicsTableAccessor::accessor_name())) {
    da->addAccessor(new PhysicsTableAccessor(fSession,*this));
  }

  if(!fRunManager) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::initialize :"
        << " no G4RunManager."
        << std::endl;   
    return false;
  } 

  if(!da->findAccessor(TrajectoryAccessor::accessor_name())) {
    da->addAccessor(new TrajectoryAccessor(fSession,fRunManager));
  }
  if(!da->findAccessor(TrajectoryContainerAccessor::accessor_name())) {
    da->addAccessor(new TrajectoryContainerAccessor(fSession,fRunManager));
  }
  if(!da->findAccessor("HC")) {
    //Declare the generic HitsCollection type :
    da->addAccessor(new HitsCollectionAccessor(fSession,fRunManager));
  }

  // Declare HitsCollections types :
  if(!fRunManager->GetUserDetectorConstruction()) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::initialize :"
        << " G4RunManager did not receive yet the"
        << " user detector construction class."
        << std::endl;   
    out << "G4Lab::Manager::initialize :"
        << " We can't then build the hits collection types."
        << std::endl;   
    return true;
  }

  G4SDManager* sdManager = G4SDManager::GetSDMpointer();
  if(!sdManager) return false;
  G4HCtable* hcTable  = sdManager->GetHCtable();
  if(!hcTable) return false;

  int number = hcTable->entries();
  for(int index=0;index<number;index++) {
    G4String hcName = hcTable->GetHCname(index);
    if(!da->findAccessor(hcName)) {
      da->addAccessor(new HitsCollectionAccessor(fSession,fRunManager,hcName));
      if(fSession.verboseLevel()) {
        std::ostream& out = fSession.cout();
        out << "G4Lab::Manager::initialize :"
            << " declare HitsCollection type " << Slash::sout(hcName)
            << std::endl;
      }
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::addAccessor(
 Slash::Data::IAccessor* aAccessor
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Data::IProcessor* da = Slash::accessorManager(fSession);
  if(!da) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::addAccessor :"
	<< " AccessorManager not found." << std::endl;
    return;
  }
  da->addAccessor(aAccessor);
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::executeG4Script(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fScriptManager) return;
  Slash::UI::IInterpreter::Options params;
  fScriptManager->executeScript("G4",aString,params);
}
//////////////////////////////////////////////////////////////////////////////
std::string G4Lab::Manager::completeCommand(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fUIsession) return "";
  return fUIsession->completeCommand(aString);
}
//////////////////////////////////////////////////////////////////////////////
bool G4Lab::Manager::executeScript(
 const std::string& aString
,const Slash::UI::IInterpreter::Options&
,const Slash::UI::IInterpreter::Aliases& aAliases
,void*
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4UImanager* UI = G4UImanager::GetUIpointer();
  if(!UI) return false;
  std::vector<std::string> text;
  Lib::smanip::lines(aString,text);
  int linen = text.size();
  if(linen) {
    for(int count=0;count<linen;count++) {
      std::string line = text[count];
      Lib::smanip::strip(line);
      if( (line[0]=='\0') || (line[0]=='#') ) continue;
      UI->ApplyCommand(line);
    }
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool G4Lab::Manager::completer(
 const std::string& aString
,std::string& aValue
,void* //aTag
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aValue = Complete(aString);
  return true;
}
#include <G4TransportationManager.hh>
#include <G4Navigator.hh>
#include <G4ParticleTable.hh>
//////////////////////////////////////////////////////////////////////////////
std::string G4Lab::Manager::physicalVolumes(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#define UNLIMITED (-1)
  G4TransportationManager* tsp = 
    G4TransportationManager::GetTransportationManager();
  if(!tsp) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::physicalVolumes : " 
        << "can't find the G4TransportationManager."
        << std::endl;
    return "";
  }
  G4Navigator* nav = tsp->GetNavigatorForTracking();
  if(!nav) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::physicalVolumes : " 
        << "can't get the G4Navigator on G4TransportationManager." 
        << std::endl;
    return "";
  }
  G4VPhysicalVolume* topVolume = nav->GetWorldVolume(); 
  if(!topVolume) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::physicalVolumes : " 
        << "no world volume in G4Navigator." << std::endl;
    return "";
  }
  std::string sout;
  sout += "<tree>";
  XML_VisitedVolume xmlVisitedVolume(sout);
  GeometryVisitor geometryVisitor;
  geometryVisitor.visit(topVolume,
                        UNLIMITED,
                        G4Transform3D(),
                        &xmlVisitedVolume);
  sout += "</tree>";
  return sout;
}
//////////////////////////////////////////////////////////////////////////////
std::string G4Lab::Manager::hitsCollections(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4SDManager* sdManager = G4SDManager::GetSDMpointer();
  if(!sdManager) return "";
  G4HCtable* hcTable  = sdManager->GetHCtable();
  if(!hcTable) return "";
  std::string sout;
  sout += "<tree>";
  int number = hcTable->entries();
  for(int index=0;index<number;index++) {
    G4String hcName = hcTable->GetHCname(index);
    sout += "<treeItem><label>";
    sout += hcName;
    sout += "</label></treeItem>";
  }
  sout += "</tree>";
  return sout;
}
//////////////////////////////////////////////////////////////////////////////
std::string G4Lab::Manager::digitsCollections(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4DigiManager* digiManager = G4DigiManager::GetDMpointer();
  if(!digiManager) return "";
  G4DCtable* dcTable  = digiManager->GetDCtable();
  if(!dcTable) return "";
  std::string sout;
  sout += "<tree>";
  int number = dcTable->entries();
  for(int index=0;index<number;index++) {
    G4String dcName = dcTable->GetDCname(index);
    sout += "<treeItem><label>";
    sout += dcName;
    sout += "</label></treeItem>";
  }
  sout += "</tree>";
  return sout;
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::setRunBeginScript(
 const std::string& aInterp
,const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fState->setRunBeginScript(aInterp,aScript);
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::setRunEndScript(
 const std::string& aInterp
,const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fState->setRunEndScript(aInterp,aScript);
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::setEventBeginScript(
 const std::string& aInterp
,const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fState->setEventBeginScript(aInterp,aScript);
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::setEventEndScript(
 const std::string& aInterp
,const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fState->setEventEndScript(aInterp,aScript);
}
//////////////////////////////////////////////////////////////////////////////
std::string G4Lab::Manager::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
void* G4Lab::Manager::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(G4Lab::Manager)
  else if_Lib_SCast(Slash::Core::IManager)
  else if_Lib_SCast(Slash::Core::IParameterManager)
  else if_Lib_SCast(IGeant4Manager)
  else return 0;
}
//////////////////////////////////////////////////////////////////////////////
G4RunManager* G4Lab::Manager::runManager(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fRunManager;
}
//////////////////////////////////////////////////////////////////////////////
G4SDManager* G4Lab::Manager::sdManager(
) const
//////////////////////////////////////////////////////////////////////////////
// Used by callbacks if having not shared libs for Geant4.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return G4SDManager::GetSDMpointer();
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::beamOn(
 int aNumber
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fRunManager) return;
  fRunManager->BeamOn(aNumber);
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::abortRun(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fRunManager) return;
  fRunManager->AbortRun();
}
//////////////////////////////////////////////////////////////////////////////
bool G4Lab::Manager::isRunning(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fState->isRunning();
}
//////////////////////////////////////////////////////////////////////////////
void G4Lab::Manager::particleTableDump(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4ParticleTable* table = G4ParticleTable::GetParticleTable();
  if(!table) return;
  table->DumpTable("ALL");
}

bool G4Lab::Manager::setParameter(const std::string&,const std::string&) {
  return false;
}

bool G4Lab::Manager::parameterValue(const std::string& aParameter,std::string& aValue) {
  if(aParameter=="physical_volumes") {
    aValue = physicalVolumes();
  } else if(aParameter=="hits_collections") {
    aValue = hitsCollections();
  } else if(aParameter=="digits_collections") {
    aValue = digitsCollections();
  } else {
    std::ostream& out = fSession.cout();
    out << "G4Lab::Manager::parameterValue :" 
        << " parameter " << Slash::sout(aParameter) << " unknown." 
        << std::endl;
    return false;
  }
  return true;  
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// From G4VBasicShell (tag geant4-05-00-ref-01)
//////////////////////////////////////////////////////////////////////////////
static G4String Complete(
 G4String commandName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4String rawCommandLine = commandName;
  G4String commandLine = rawCommandLine.strip(G4String::both);
  size_t i = commandLine.index(" ");
  if( i != std::string::npos ) return rawCommandLine; // Already entering parameters, 
                                            // assume command path is correct.
  G4String commandString = commandLine; 
  //G4String targetCom = ModifyPath(commandString);
  G4String targetCom = commandString;
  G4UIcommandTree* tree = G4UImanager::GetUIpointer()->GetTree();
  G4String value = FindMatchingPath(tree,targetCom);
  if(value.empty()) return rawCommandLine;
  return value;
}
//////////////////////////////////////////////////////////////////////////////
static G4String FindMatchingPath(
 G4UIcommandTree* aTree
,G4String aCommandPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
// From intercoms/src/G4UIcommandTree::FindPath.
{
  G4String empty = "";
  if(aTree==NULL) return empty;
  G4String pathName = aTree->GetPathName();
  if( aCommandPath.index( pathName ) == std::string::npos ) return empty;
  G4String remainingPath = aCommandPath;
  remainingPath.remove(0,pathName.length());
  size_t i = remainingPath.first('/');
  if( i == std::string::npos ) {
    // Look for number of matching commands :
    std::vector<G4UIcommand*> commands;
    G4int n_commandEntry = aTree->GetCommandEntry();
    for( G4int i_thCommand = 1; i_thCommand <= n_commandEntry; i_thCommand++ ) {
      G4UIcommand* cmd = aTree->GetCommand(i_thCommand);
      G4String ss = cmd->GetCommandName();
      ss.resize(remainingPath.length());
      if( remainingPath == ss ) commands.push_back(cmd);
    }
    n_commandEntry = commands.size();
    if(n_commandEntry==1) {
      return (pathName + commands[0]->GetCommandName());
    } else if (n_commandEntry>=2) {
      G4cout << "Matching commands :" << G4endl; 
      for( G4int i_thCommand = 0; i_thCommand < n_commandEntry; i_thCommand++ ) {
        G4UIcommand* cmd = commands[i_thCommand];
        G4cout << cmd->GetCommandName() << G4endl; 
      }
      return empty;
    }
    // Look for sub tree :
    std::vector<G4UIcommandTree*> trees;
    G4String nextPath = pathName;
    nextPath.append(remainingPath);
    G4int n_treeEntry = aTree->GetTreeEntry();
    for( G4int i_thTree = 1; i_thTree <= n_treeEntry; i_thTree++ ) {
      G4UIcommandTree* tree = aTree->GetTree(i_thTree);
      G4String ss = tree->GetPathName();
      ss.resize(nextPath.length());
      if( nextPath == ss ) trees.push_back(tree);
    }
    n_treeEntry = trees.size();
    if(n_treeEntry==1) {
      return trees[0]->GetPathName();
    } else if (n_treeEntry>=2) {
      G4cout << "Matching directories :" << G4endl; 
      for( G4int i_thTree = 0; i_thTree < n_treeEntry; i_thTree++ ) {
        G4UIcommandTree* tree = trees[i_thTree];
        G4cout << tree->GetPathName() << G4endl; 
      }
      return empty;
    } else {
      return empty; // No match.
    }
  } else {
    // Find path
    G4String nextPath = pathName;
    nextPath.append(remainingPath(0,i+1));
    G4int n_treeEntry = aTree->GetTreeEntry();
    for( G4int i_thTree = 1; i_thTree <= n_treeEntry; i_thTree++ ) {
      G4UIcommandTree* tree = aTree->GetTree(i_thTree);
      if( nextPath == tree->GetPathName() ) { 
        return FindMatchingPath(tree,aCommandPath ); 
      }
    }
  }
  return empty;
}

#include "G4LossTableManager.hh"
void G4Lab::Manager::verboseOff(){
  G4LossTableManager::Instance()->SetVerbose(-1);
}
