// this :
#include <G4Lab/PhysicsTableAccessor.h>

// Geant4 :
#include <G4ParticleTable.hh>
#include <G4PhysicsTable.hh>
#include <G4ParticleDefinition.hh>
#include <G4Element.hh>
#include <G4VProcess.hh>
#include <G4ProcessVector.hh>
#include <G4ProcessManager.hh>

#include <Slash/Core/ISession.h>
#include <Slash/Data/IIterator.h>
#include <Slash/Tools/smanip.h>
#include <Slash/Tools/fmanip.h>
#include <Slash/Tools/directory.h>
#include <Slash/Tools/sprintf.h>

// Lib :
#include <Lib/Value.h>
#include <Lib/Cast.h>
#include <Lib/Debug.h>

// AIDA :
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <AIDA/IPlotterFactory.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud2D.h>

// G4Lab :
#include <G4Lab/Interfaces/IGeant4Manager.h>

namespace G4Lab {
class PhysicsTable {
public:
  PhysicsTable(const std::string& aTableName,
               const std::string& aProcessName,
               const std::string& aParticleName,
               G4PhysicsTable* aTable,
               G4VProcess* aProcess,
               G4ParticleDefinition* aParticle)
    :fTableName(aTableName)
    ,fProcessName(aProcessName)
    ,fParticleName(aParticleName)
    ,fTable(aTable)
    ,fProcess(aProcess)
    ,fParticle(aParticle){}
  virtual ~PhysicsTable() {
    delete fTable;
  }
public:
  std::string fTableName;
  std::string fProcessName;
  std::string fParticleName;
  G4PhysicsTable* fTable;
  G4VProcess* fProcess;
  G4ParticleDefinition* fParticle;
  Lib::Debug fDebug;
};
}

std::string G4Lab::PhysicsTableAccessor::accessor_name() {return "PhysicsTable";}

G4Lab::PhysicsTableAccessor::PhysicsTableAccessor(
 Slash::Core::ISession& aSession
,IGeant4Manager& aManager
)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fManager(aManager)
,fType(accessor_name())
,fAIDA(0)
{
  addProperty("table",Lib::Property::STRING);
  addProperty("process",Lib::Property::STRING);
  addProperty("particle",Lib::Property::STRING);
  addProperty("number",Lib::Property::INTEGER);
  addProperty("name",Lib::Property::STRING);
}

G4Lab::PhysicsTableAccessor::~PhysicsTableAccessor() {
  for(unsigned int index=0;index<fTables.size();index++) delete fTables[index];
}

void* G4Lab::PhysicsTableAccessor::cast(const std::string& aClass) const {
  if_Lib_SCast(Slash::Data::IVisualizer)
  else return Lib::BaseAccessor::cast(aClass);
}

std::string G4Lab::PhysicsTableAccessor::name() const {return fType;}

namespace G4Lab {
  class PhysicsTableIterator : public virtual Slash::Data::IIterator {
  public: //Slash::Data::IIterator
    virtual Slash::Data::IAccessor::Data object() {
      if(fIterator==fVector.end()) return 0;
      return *fIterator;
    }
    virtual void next() { ++fIterator; }
    virtual void* tag() { return 0;}
  public:
    PhysicsTableIterator(std::vector<G4Lab::PhysicsTable*>& aVector)
      :fVector(aVector) {
      fIterator = fVector.begin();
    }
    virtual ~PhysicsTableIterator() {}
  private:
    std::vector<G4Lab::PhysicsTable*>& fVector;
    std::vector<G4Lab::PhysicsTable*>::iterator fIterator;
  };
}

Slash::Data::IIterator* G4Lab::PhysicsTableAccessor::iterator() {
  if(!fTables.size()) {
    if(!buildProcessTables()) return 0;
  }
  return new PhysicsTableIterator(fTables);
}

Slash::Core::IValue* G4Lab::PhysicsTableAccessor::findValue(
 Slash::Data::IAccessor::Data aData
,const std::string& aName
,void*
){
  PhysicsTable* obj = (PhysicsTable*)aData;
  if(aName=="table") {
    return new Lib::Value(obj->fTableName);
  } else if(aName=="process") {
    return new Lib::Value(obj->fProcessName);
  } else if(aName=="particle") {
    return new Lib::Value(obj->fParticleName);
  } else if(aName=="number") {
    return new Lib::Value((int)obj->fTable->size());
  } else if(aName=="name") {
    std::string name = 
      obj->fTableName + "." + obj->fProcessName + "." + obj->fParticleName;
    return new Lib::Value(name);
  } else {
    return new Lib::Value();
  }
}

#include "find_AIDA.cxx"

void G4Lab::PhysicsTableAccessor::beginVisualize(){
  if(!fAIDA) fAIDA = find_AIDA(fSession);
  if(!fAIDA) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::PhysicsTableAccessor::beginVisualize :"
        << " AIDA not found. Can't then find an AIDA plotter." 
        << std::endl;
  }
}

void G4Lab::PhysicsTableAccessor::visualize(
 Slash::Data::IAccessor::Data aData
,void*
){
  PhysicsTable* obj = (PhysicsTable*)aData;

  std::string value;

  bool superpose = true;
  //if(fSession.parameterValue("modeling.superpose",value))
  //FIXME : avoid to have the name "OnXLab_tree" here.
  if(fSession.parameterValue("OnXLab_tree.append",value))
    if(!Slash::to(value,superpose)) superpose = true;

  if(!fAIDA) return;
    
  AIDA::ITreeFactory* tf = fAIDA->createTreeFactory();
  if(!tf) return;
  AIDA::ITree* memoryTree = tf->create();
  delete tf;
  if(!memoryTree) return;

  AIDA::IPlotterFactory* plotterFactory = fAIDA->createPlotterFactory();
  if(!plotterFactory) {
    std::ostream& out = fSession.cout();
    out << "Can't create a plotter factory." << std::endl;
    return;
  }

  AIDA::IPlotter* plotter = plotterFactory->create();
  delete plotterFactory;
  if(!plotter) {
    std::ostream& out = fSession.cout();
    out << "Can't create a plotter factory." << std::endl;
    return;
  }

  AIDA::IHistogramFactory* histogramFactory = 
    fAIDA->createHistogramFactory(*memoryTree);
  if(!histogramFactory) {
    std::ostream& out = fSession.cout();
    out << "Can't create an histogram factory." << std::endl;
    delete plotter;
    return;
  }

  int number = obj->fTable->entries();
  if(!superpose) {
    if(number>0) plotter->createRegions(1,number,0);
    else plotter->createRegions(1,1,0);
  }

  for (int count=0;count<number;count++){
    G4PhysicsVector* physicsVector = (*(obj->fTable))[count];
    size_t n = physicsVector->GetVectorLength();
    if(n<2) {
      std::ostream& out = fSession.cout();
      out << "PhysicsVector with " << int(n) << " entries. Can't plot." 
          << std::endl;
      continue;
    }

    std::string name;
    name += obj->fTableName;
    name += ".";
    name += obj->fProcessName;
    name += ".";
    name += obj->fParticleName;
    name += ".";
    name += Slash::to<int>(count);
    AIDA::ICloud2D* cloud = 
      histogramFactory->createCloud2D(name,name);
    if(!cloud) {
      std::ostream& out = fSession.cout();
      out << "Can't create \"" << name << "\" cloud1D." << std::endl;
      break;
    }
    for (unsigned int i=0;i<n;i++) {
      cloud->fill(i,(*physicsVector)[i]);
    }
      
    AIDA::IPlotterRegion& region = plotter->currentRegion();

    //options (from OpenPAW/pavect.cxx)
    std::string field = "plotter.pointsStyle[0].";
#define SOLID  "0xFFFF"
#define SbModeling_lines "lines"
    region.setParameter(field+"linePattern",SOLID);
    region.setParameter(field+"modeling",SbModeling_lines);
    region.setParameter(field+"painting","UNIFORM");
    region.setParameter(field+"color","0 0 0");

    region.plot(*cloud);
    /*
    region.setParameter("plotter.xAxisAutomated","FALSE");
    region.setParameter("plotter.xAxisMinimum",Slash::to<double>(0));
    region.setParameter("plotter.xAxisMaximum",Slash::to<double>(n-1));
    region.setParameter("plotter.xAxisMaximum",Slash::to<double>(n-1));
    //region.setParameter("plotter.xNumberOfPoints",Slash::to<int>(4*n));
    */
    /*
    fSession.out().println("debug : min : %s, max : %s, steps : %s\n.",
                           value1.c_str(),value2.c_str(),value3.c_str());
    */
    if(!superpose) plotter->next();
      
  }
  
  delete histogramFactory;
  delete plotter;
    
}

void G4Lab::PhysicsTableAccessor::endVisualize() {}

bool G4Lab::PhysicsTableAccessor::buildProcessTables() {
  if(fTables.size()) return true; //done.

  if(!fManager.isRunning()) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
        << " It is needed to have started a run to get physics tables." 
        << std::endl;
    return false;
  }

  G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
  if(!particleTable) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
        << " Can't get particle table." 
        << std::endl;
    return false;
  }

  G4ParticleTable::G4PTblDicIterator* particleTableIterator = 
    particleTable->GetIterator();
  if(!particleTableIterator) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
        << " Can't get particle table iterator." 
        << std::endl;
    return false;
  }

  particleTableIterator->reset();
  while( (*particleTableIterator)() ){
    G4ParticleDefinition* particle = particleTableIterator->value();
    if(!particle) {
      std::ostream& out = fSession.cout();
      out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
          << " Can't get particle table particle." 
          << std::endl;
      return false;
    }
    G4String particleName = particle->GetParticleName();
    //std::ostream& out = fSession.cout();
    //out << "debug : Particle : " << particleName << std::endl;

    // Processes for this particle :
    G4ProcessManager* processManager = particle->GetProcessManager();
    if(!processManager) {
      std::ostream& out = fSession.cout();
      out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
          << " Can't get process manager of particle " 
          << particleName << "."
          << std::endl;
      return false;
    }

    G4ProcessVector* processList = processManager->GetProcessList();
    if(!processList) {
      std::ostream& out = fSession.cout();
      out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
          << " Can't get process list for process manager of particle " 
          << particleName << "."
          << std::endl;
      return false;
    }

    int number = processList->entries();
    for (int index=0;index<number;index++){
      G4VProcess* process = (*processList)(index);
      std::vector<std::string> names;
      std::vector<G4PhysicsTable*> tables;
      if(!findProcessTables(particle,process,names,tables)) {
        return false;
      }

      for(unsigned int i=0;i<tables.size();i++) {
        std::vector<std::string> words;
        Slash::words(names[i],".",false,words);
        if(words.size()!=3) {
          std::ostream& out = fSession.cout();
          out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
              << " Three words separted with a dot expected in " 
              << names[i] << " for particle " << particleName << "."
              << std::endl;
          return false;
        }
        fTables.push_back(new PhysicsTable(words[0],
                                           words[1],
                                           words[2],
                                           tables[i],
                                           process,
                                           particle));
      }
    }
  }

  return true;
}

bool G4Lab::PhysicsTableAccessor::findProcessTables(
 G4ParticleDefinition* aParticle
,G4VProcess* aProcess
,std::vector<std::string>& aTables
,std::vector<G4PhysicsTable*>& aPhysicsTables
){
  aTables.clear();
  aPhysicsTables.clear();

  std::string tmpName;
  if(!Slash::tmpname(".","G4Lab_tmp","",tmpName)) {
    std::ostream& out = fSession.cout();
    out << "Can't build a temporary directory name." << std::endl;
    return false;
  }

  if(!Slash::directory::create(tmpName)) {
    std::ostream& out = fSession.cout();
    out << "Can't create \"" << tmpName << "\" directory." << std::endl;
    return false;
  }

#ifdef WIN32 //binary mode out on Windows.
  G4bool ascii = true;
#else
  G4bool ascii = false;
  //G4bool ascii = true;
#endif

  if(!aProcess->StorePhysicsTable(aParticle,tmpName,ascii)) {
    std::ostream& out = fSession.cout();
    out << "Can't store physics table for particle "
        << aParticle->GetParticleName()
        << " and process " << aProcess->GetProcessName() << "."
        << std::endl;
    //return false;
  } else {

    // Analyse the content of the directory :
    std::vector<std::string> files;
    if(!Slash::directory::entries(tmpName,files,false)) {
      std::ostream& out = fSession.cout();
      out << "Can't get files in \"" << tmpName << "\" directory." 
          << std::endl;
    } else {
      // Analyse files :
      Slash::remove(files,".");
      Slash::remove(files,"..");
/*
      if(!files.size()) {
        std::ostream& out = fSession.cout();
        out << "No files in \"" << tmpName << "\" directory for particle " 
            << aParticle->GetParticleName() 
            << " and process " << aProcess->GetProcessName() << "."
            << std::endl;
        //return false;
      }
*/
      for(unsigned int index=0;index<files.size();index++) {
        std::vector<std::string> words;
        Slash::words(files[index],".",false,words);
        //std::ostream& out = fSession.cout();
        //out << "debug : Analyse \"" << files[index] << "\" ." << std::endl;
        if(words.size()!=4) {
          std::ostream& out = fSession.cout();
          out << "Bad syntax in file name " << Slash::sout(files[index])
              << " for particle " << aParticle->GetParticleName() 
              << " and process " << aProcess->GetProcessName() << "."
              << std::endl;
          //aTables.clear();
          //aPhysicsTables.clear();
          //return false;
          continue;
        }
        std::string table = words[0];
        std::string process = words[1];
        std::string particle = words[2];
        //Slash::replace(aTables[i],".dat","");
        std::string filename = 
          aProcess->GetPhysicsTableFileName(aParticle,
                                            tmpName,
                                            table,
                                            ascii);
        G4PhysicsTable* physicsTable = 
          new G4PhysicsTable(G4Element::GetNumberOfElements());
        G4bool status = physicsTable->RetrievePhysicsTable(filename,ascii);
        if(!status){
          std::ostream& out = fSession.cout();
          out << "Unable to read \"" << filename << "\" file." << std::endl;
          delete physicsTable;
          physicsTable = 0;
        }
        if(physicsTable) {
          aTables.push_back(table+"."+process+"."+particle);
          aPhysicsTables.push_back(physicsTable);
        }
      }
    }

  }

  // Remove the temporary directory :
  if(!Slash::directory::rmdir(tmpName)) {
    std::ostream& out = fSession.cout();
    out << "Can't remove \"" << tmpName << "\" directory." << std::endl;
    aTables.clear();
    aPhysicsTables.clear();
    return false;
  }

  return true;
}
