/*
 * AIDA analysis program over G4LabTemplateTwo.root file.
 */

// AIDA :
#include <AIDA/AIDA.h>

// std :
#include <iostream>

#include <math.h>

static void computeError(AIDA::IHistogram1D&,AIDA::IDataPointSet&);
static void set_region_style(AIDA::IPlotterRegion&);
static bool plot_pi_plus(AIDA::IPlotterRegion&,const std::vector<AIDA::IDataPointSet*>&,AIDA::IHistogramFactory&);
static bool plot_pi_minus(AIDA::IPlotterRegion&,const std::vector<AIDA::IDataPointSet*>&,AIDA::IHistogramFactory&);
static bool plot_K_plus(AIDA::IPlotterRegion&,const std::vector<AIDA::IDataPointSet*>&,AIDA::IHistogramFactory&);
static bool plot_K_minus(AIDA::IPlotterRegion&,const std::vector<AIDA::IDataPointSet*>&,AIDA::IHistogramFactory&);

const int numberOfXFValues = 4;
const double nProton = 1e6;        // number of incident protons

//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string arg1 = aArgc>=2?aArgv[1]:"";

  std::string program = "G4LabTemplateTwo_analysis";
  std::string header = program + " : ";

  AIDA::IAnalysisFactory* aida = AIDA_createAnalysisFactory();
  if(!aida) {
    std::cout << header << "AIDA not found." << std::endl;
    return 1;
  }

  AIDA::ITreeFactory* treeFactory = aida->createTreeFactory();
  if(!treeFactory) {
    std::cout << header << "can't get a TreeFactory." << std::endl;
    return 1;
  }

  AIDA::ITree* tree = treeFactory->create("G4LabTemplateTwo.root","root",true,false);
  if(!tree) {
    std::cout << header << "can't open data file." << std::endl;
    return 1;
  }

  AIDA::IManagedObject* object = tree->find("tupleOFFTarget");
  if(!object) {
    std::cout << header 
              << "object tupleOFFTarget not found in tree." 
              << std::endl;
    return 1;
  }
  AIDA::ITuple* tuple = dynamic_cast<AIDA::ITuple*>(object);
  if(!tuple) {
    std::cout << header << "object not an AIDA::ITuple." << std::endl;
    return 1;
  }

  int coln = tuple->columns();
  for(int index=0;index<coln;index++) {
    std::cout << header 
              << "column " << tuple->columnName(index) 
              << " type " << tuple->columnType(index) 
              << std::endl;
  }

  std::cout << header << "rows = " << tuple->rows() << std::endl;

  int numberOfParticles = 4;

  // Create histograms in memory tree :
  AIDA::ITree* memory = treeFactory->create();
  if(!tuple) {
    std::cout << header << "can't get memory tree." << std::endl;
    return 1;
  }
  AIDA::IHistogramFactory* histogramFactory = 
    aida->createHistogramFactory(*memory);
  if(!histogramFactory) {
    std::cout << header << "can't get an histogram factory." << std::endl;
    return 1;
  }
  AIDA::IDataPointSetFactory* dpsFactory = 
    aida->createDataPointSetFactory(*memory);
  if(!dpsFactory) {
    std::cout << header << "can't get an data point set factory." << std::endl;
    return 1;
  }

  std::vector< std::vector<AIDA::IHistogram1D*> > histos;
  histos.resize(numberOfParticles);
  std::vector< std::vector<AIDA::IDataPointSet*> > dpss;
  dpss.resize(numberOfParticles);
  int ipart; 
  for (ipart=0; ipart<numberOfParticles; ++ipart) {
    histos[ipart].resize(numberOfXFValues);
    dpss[ipart].resize(numberOfXFValues);
  }

  const char* hxfValue[] = {"0.034","0.15","0.50","0.75"}; //values of the xF
  const char* hpartName[] = {"pip","pim","Kp","Km"}; //particle Name
  int nPTBins = 10;           // number of bins in pT
  double pTMin = 0.;           // pTmin (GeV)
  double pTMax = 0.6;          // pTMax (GeV)

  int ixf; 
  for (ipart=0; ipart<numberOfParticles; ++ipart) { //loop on particle types
    for (ixf=0; ixf<numberOfXFValues; ++ixf) { //loop on xF ranges
      std::string title1 = "E ds/dp (b/GeV2) xF=";
      title1 += hxfValue[ixf];
      title1 += hpartName[ipart];
      std::string id1 = "_1_";
      id1 += hpartName[ipart];
      id1 += hxfValue[ixf];
      histos[ipart][ixf] = 
        histogramFactory->createHistogram1D(id1,title1,nPTBins,pTMin,pTMax);

      std::string title2 = " DPS xF=";
      title2 += hxfValue[ixf];
      title2 += hpartName[ipart];
      std::string id2 = "_2_";
      id2 += hpartName[ipart];
      id2 += hxfValue[ixf];
      dpss[ipart][ixf] = dpsFactory->create(id2,title2,2);

    }//xF loop
  }//particle loop

  double* xFCentral = new double[numberOfXFValues];
  double* xFMin = new double[numberOfXFValues];
  double* xFMax = new double[numberOfXFValues];   // [Min,Max] range in xF 
  for (ixf=0 ; ixf<numberOfXFValues ; ++ixf) { //define the xF ranges
    xFCentral[ixf] = std::atof(hxfValue[ixf]);
    xFMin[ixf] = 0.90 * xFCentral[ixf];
    xFMax[ixf] = 1.10 * xFCentral[ixf];
    std::cout << header
              << "xFMin["<<ixf<<"]=" << xFMin[ixf] 
              << " "
              << "xFMax["<<ixf<<"]=" << xFMax[ixf] 
              << std::endl;
  }


  double pionMass2 = ::pow(0.1396,2.0);             // m_pi^2
  double twopi= 6.28318;       // 2pi
  double Emax = 400.0;         // maximum beam energy (momentum) GeV/c
  
  double pTBinWidth=(pTMax-pTMin)/nPTBins; // Bin Width in pT and pL
  double targetFactor = 0.989; // A/(rho. Navogadro . Length_equiv) [barn]

  //int nentries = tuple->rows();
  int nentries = 100000;
  std::cout << header 
            << "traitements de " << nentries << " entrees" 
            << std::endl;

  int irow = 0;
  tuple->start();
  while(tuple->next() && (irow<nentries)) {
    int pdg = tuple->getInt(0);
    /*double x =*/ tuple->getDouble(1);
    /*double y =*/ tuple->getDouble(2);
    /*double z =*/ tuple->getDouble(3);
    double px = tuple->getDouble(4);
    double py = tuple->getDouble(5);
    double pz = tuple->getDouble(6);

    
    double pL = pz;
    double pT = ::sqrt(px*px + py*py);
    double ptot = ::sqrt(pL*pL + pT*pT);
    double Etot = ::sqrt(ptot*ptot + pionMass2);
    double xF = pL/Emax;
    double weight1 = (Etot/pT);
    weight1 = weight1/twopi;
    weight1 = weight1 * targetFactor/nProton;
    weight1 = weight1 / pTBinWidth;
    /*
     *pdg code:
     *	 p    : 2212
     *	 pbar: -2212
     *	 n    : 2112
     *	 pi+:    211
     *	 pi-:   -211
     *	 pi0:    111
     *	 K+:     321
     *	 K-:    -321
     *	 K0S:    310
     *	 K0L:    130
     *	 e+:     -11
     *	 e-:     +11
     *	 mu+:    -13
     *	 mu-:    +13
     *        gamma:   22
     */
    for(ixf=0 ; ixf<numberOfXFValues ; ++ixf) {
      double pLBinWidth = Emax * (xFMax[ixf] - xFMin[ixf]);
      double weight = weight1/pLBinWidth;
      if ( (xF >= xFMin[ixf]) && (xF <= xFMax[ixf]) ) {
        if ( fabs(211.0-pdg) < 0.1) {          // pi+
          histos[0][ixf]->fill(pT,weight);
        } else if ( fabs(-211.0-pdg) < 0.1 ) { // pi-
          histos[1][ixf]->fill(pT,weight);
        } else if ( fabs(321.0-pdg) < 0.1 ) {  // K+
          histos[2][ixf]->fill(pT,weight);
        } else if ( fabs(-321.0-pdg) < 0.1 ) { // K-
          histos[3][ixf]->fill(pT,weight);
        } // pdg code switch
      } //xF range x-check
    } //xF predifined values

    irow++;
  }

  //Recompute all the error bin
  for (ipart=0; ipart<numberOfParticles; ++ipart) { //loop on particle types
    for (ixf=0; ixf<numberOfXFValues; ++ixf) { //loop on xF ranges
      computeError(*histos[ipart][ixf],*dpss[ipart][ixf]);
    }
  }
  
  //Plotting
  AIDA::IPlotterFactory* plotterFactory = 
    aida->createPlotterFactory(aArgc,aArgv);
  if(!plotterFactory) {
    std::cout << header << "can't get a PlotterFactory." << std::endl;
    return 1;
  }
    
  AIDA::IPlotter* plotter = plotterFactory->create();
  if(!plotter) {
    std::cout << header << "can't get a plotter." << std::endl;
    return 1;
  }
    
  plotter->clearRegions();
  bool debug = arg1=="-debug"?true:false;
  //plotter->createRegions(1,numberOfParticles,0);
  if(debug) plotter->createRegions(1,1,0);
  else      plotter->createRegions(2,2,0);

  //FIXME : set size to 800 600
  //FIXME : set Y to log.
  //FIXME : set PS production (mode 111 in ROOT ???)
  //static const char* commonFileName = "G4LabTemplateTwo.ps";

  plotter->setCurrentRegionNumber(0);
  plot_pi_plus(plotter->currentRegion(),dpss[0],*histogramFactory);
  if(!debug) {
  plotter->setCurrentRegionNumber(1);
  plot_pi_minus(plotter->currentRegion(),dpss[1],*histogramFactory);

  plotter->setCurrentRegionNumber(2);
  plot_K_plus(plotter->currentRegion(),dpss[2],*histogramFactory);

  plotter->setCurrentRegionNumber(3);
  plot_K_minus(plotter->currentRegion(),dpss[3],*histogramFactory);
  }

  //plotter.refresh();
  //FIXME : explicitly close file.

  plotter->show();
  plotter->interact();
  delete plotter;
  delete plotterFactory;

  delete [] xFCentral;
  delete [] xFMin;
  delete [] xFMax;

  delete dpsFactory;
  delete histogramFactory;
  delete treeFactory;
  delete tree;
  delete aida;

  return 0;
}

void computeError(AIDA::IHistogram1D& aHisto,AIDA::IDataPointSet& aDPS) {
  int binn = aHisto.axis().bins();
  for(int i=0;i<binn;i++) {
    int nContent = aHisto.binEntries(i);
    double w = aHisto.binHeight(i);
    double we = 0;
    if (nContent>0) we = w * ::sqrt(1./nContent - 1./nProton);
    double x = (aHisto.axis().binUpperEdge(i)+aHisto.axis().binLowerEdge(i))/2;
    double dx = aHisto.axis().binUpperEdge(i)-aHisto.axis().binLowerEdge(i);
    AIDA::IDataPoint* point = aDPS.addPoint();
    AIDA::IMeasurement* m;
    m = point->coordinate(0);
    m->setValue(x);
    m->setErrorMinus(dx/2);
    m->setErrorPlus(dx/2);
    m = point->coordinate(1);
    m->setValue(w);
    m->setErrorMinus(we/2);
    m->setErrorPlus(we/2);
  }
}

// Lib :
#include <Lib/smanip.h> //FIXME : use STL ?

//////////////////////////////////////////////////////////////////////////////
bool plot_pi_plus(
 AIDA::IPlotterRegion& aRegion
,const std::vector<AIDA::IDataPointSet*>& aDPS
,AIDA::IHistogramFactory& aHistogramFactory 
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  set_region_style(aRegion);

  //xmin, ymin,xmax, ymax
  aRegion.setParameter("plotter.yAxisAutomated","FALSE");
  aRegion.setParameter("plotter.yAxisMinimum","1.e-4");
  aRegion.setParameter("plotter.yAxisMaximum","1");
  //FIXME : handle "#pi^{+}"
  aRegion.setTitle("[p]^+!");

  //  Color_t xfColor[] = {kGreen,kBlue,kRed,kBlack};
  const char* sxfColor[] = { 
    "0 1 0", //green
    "0 0 1", //blue
    "1 0 0", //red
    "0 0 0"  "black"
  };

  int ixf;
  for (ixf=0; ixf<numberOfXFValues; ++ixf) { //loop on xF ranges
    std::string sbin = "plotter.binsStyle[";
    sbin += Lib::smanip::tostring(ixf);
    sbin += "].";
    aRegion.setParameter(sbin+"visible","FALSE");
    //aRegion.setParameter(sbin+"modeling","wire_boxes");
    //aRegion.setParameter(sbin+"modeling","top_lines");
    //aRegion.setParameter(sbin+"color",sxfColor[ixf]);

    std::string serror = "plotter.errorsStyle[";
    serror += Lib::smanip::tostring(ixf);
    serror += "].";
    aRegion.setParameter(serror+"visible","TRUE");
    aRegion.setParameter(serror+"color",sxfColor[ixf]);
    aRegion.setParameter(serror+"lineWidth","2");
    aRegion.plot(*aDPS[ixf]);
  }//loop on xf values
  
  //exp data from Bonesini et al

  double xpm1[] = {0.   , 0.075  ,  0.15 , 0.225 , 0.3375, 0.45};
  double ypm1[] = {0.3479, 0.3699, 0.3341, 0.2422, 0.1277, 0.0684};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_1","");
  if(!cloud) return false;
  //FIXME : handle  : "x_{F}=0.034"
  cloud->annotation().addItem("Legend","x?F!=0.034");
  for(int index=0;index<6;index++) cloud->fill(xpm1[index],ypm1[index],1);
  aRegion.setParameter("plotter.pointsStyle[0].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[0].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[0].color",sxfColor[0]);
  aRegion.plot(*cloud);}

  double xpm2[] = {0.,     0.5};
  double ypm2[] = {0.108,  0.0338};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_2","");
  if(!cloud) return false;
  //FIXME : handle : "x_{F}=0.15"
  cloud->annotation().addItem("Legend","x?F!=0.15");
  for(int index=0;index<2;index++)cloud->fill(xpm2[index],ypm2[index],1);
  aRegion.setParameter("plotter.pointsStyle[1].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[1].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[1].color",sxfColor[1]);
  aRegion.plot(*cloud);}

  double xpm3[] = {0.,     0.5};
  double ypm3[] = {0.0146, 0.00449};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_3","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.50"
  cloud->annotation().addItem("Legend","x?F!=0.50");
  for(int index=0;index<2;index++) cloud->fill(xpm3[index],ypm3[index],1);
  aRegion.setParameter("plotter.pointsStyle[2].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[2].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[2].color",sxfColor[2]);
  aRegion.plot(*cloud);}

  double xpm4[] = {0.5};
  double ypm4[] = {0.0005};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_4","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.75"
  cloud->annotation().addItem("Legend","x?F!=0.75");
  for(int index=0;index<1;index++) cloud->fill(xpm4[index],ypm4[index],1);
  aRegion.setParameter("plotter.pointsStyle[3].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[3].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[3].color",sxfColor[3]);
  aRegion.plot(*cloud);}

  return true;
}

//////////////////////////////////////////////////////////////////////////////
bool plot_pi_minus(
 AIDA::IPlotterRegion& aRegion
,const std::vector<AIDA::IDataPointSet*>& aDPS
,AIDA::IHistogramFactory& aHistogramFactory 
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  set_region_style(aRegion);

  aRegion.setParameter("plotter.yAxisAutomated","FALSE");
  aRegion.setParameter("plotter.yAxisMinimum","1.e-4");
  aRegion.setParameter("plotter.yAxisMaximum","1");
  //FIXME : handle "#pi^{-}"
  aRegion.setTitle("[p]^-!");

  //  Color_t xfColor[] = {kGreen,kBlue,kRed,kBlack};
  const char* sxfColor[] = { 
    "0 1 0", //green
    "0 0 1", //blue
    "1 0 0", //red
    "0 0 0"  "black"
  };

  int ixf;
  for (ixf=0; ixf<numberOfXFValues; ++ixf) { //loop on xF ranges
    std::string sbin = "plotter.binsStyle[";
    sbin += Lib::smanip::tostring(ixf);
    sbin += "].";
    aRegion.setParameter(sbin+"visible","FALSE");
    //aRegion.setParameter(sbin+"modeling","wire_boxes");
    //aRegion.setParameter(sbin+"modeling","top_lines");
    //aRegion.setParameter(sbin+"color",sxfColor[ixf]);

    std::string serror = "plotter.errorsStyle[";
    serror += Lib::smanip::tostring(ixf);
    serror += "].";
    aRegion.setParameter(serror+"visible","TRUE");
    aRegion.setParameter(serror+"color",sxfColor[ixf]);
    aRegion.setParameter(serror+"lineWidth","2");
    aRegion.plot(*aDPS[ixf]);

  }//loop on xf values
  
  //exp data from Bonesini et al

  double xpm10[] = {0., 0.075,  0.225};
  double ypm10[] = {0.2892, 0.2936, 0.2001};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_10","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.034"
  cloud->annotation().addItem("Legend","x?F!=0.034");
  for(int index=0;index<3;index++) cloud->fill(xpm10[index],ypm10[index],1);
  aRegion.setParameter("plotter.pointsStyle[0].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[0].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[0].color",sxfColor[0]);
  aRegion.plot(*cloud);}

  double xpm20[] = {0, 0.5};
  double ypm20[] = {0.0675, 0.0223};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_20","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.15"
  cloud->annotation().addItem("Legend","x?F!=0.15");
  for(int index=0;index<2;index++) cloud->fill(xpm20[index],ypm20[index],1);
  aRegion.setParameter("plotter.pointsStyle[1].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[1].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[1].color",sxfColor[1]);
  aRegion.plot(*cloud);}

  double xpm30[] = {0., 0.5};
  double ypm30[] = {0.005,  0.00164};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_30","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.50"
  cloud->annotation().addItem("Legend","x?F!=0.50");
  for(int index=0;index<2;index++) cloud->fill(xpm30[index],ypm30[index],1);
  aRegion.setParameter("plotter.pointsStyle[2].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[2].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[2].color",sxfColor[2]);
  aRegion.plot(*cloud);}

  double xpm40[] = {0., 0.5};
  double ypm40[] = {0.00027, 0.000114};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_40","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.75"
  cloud->annotation().addItem("Legend","x?F!=0.75");
  for(int index=0;index<2;index++) cloud->fill(xpm40[index],ypm40[index],1);
  aRegion.setParameter("plotter.pointsStyle[3].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[3].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[3].color",sxfColor[3]);
  aRegion.plot(*cloud);}

  return true;
}


//////////////////////////////////////////////////////////////////////////////
bool plot_K_plus(
 AIDA::IPlotterRegion& aRegion
,const std::vector<AIDA::IDataPointSet*>& aDPS
,AIDA::IHistogramFactory& aHistogramFactory 
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  set_region_style(aRegion);

  aRegion.setParameter("plotter.yAxisAutomated","FALSE");
  aRegion.setParameter("plotter.yAxisMinimum","1.e-4");
  aRegion.setParameter("plotter.yAxisMaximum","1.e-1");
  //FIXME : handle "#K^{+}"
  aRegion.setTitle("K^+!");

  //  Color_t xfColor[] = {kGreen,kBlue,kRed,kBlack};
  const char* sxfColor[] = { 
    "0 1 0", //green
    "0 0 1", //blue
    "1 0 0", //red
    "0 0 0"  "black"
  };

  int ixf;
  for (ixf=0; ixf<numberOfXFValues; ++ixf) { //loop on xF ranges
    std::string sbin = "plotter.binsStyle[";
    sbin += Lib::smanip::tostring(ixf);
    sbin += "].";
    aRegion.setParameter(sbin+"visible","FALSE");
    //aRegion.setParameter(sbin+"modeling","wire_boxes");
    //aRegion.setParameter(sbin+"modeling","top_lines");
    //aRegion.setParameter(sbin+"color",sxfColor[ixf]);

    std::string serror = "plotter.errorsStyle[";
    serror += Lib::smanip::tostring(ixf);
    serror += "].";
    aRegion.setParameter(serror+"visible","TRUE");
    aRegion.setParameter(serror+"color",sxfColor[ixf]);
    aRegion.setParameter(serror+"lineWidth","2");
    aRegion.plot(*aDPS[ixf]);
  }//loop on xf values
  
  //exp data from Bonesini et al

  double xpm11[] = {0., 0.075,  0.15,   0.225,  0.3375, 0.45};
  double ypm11[] = {0.0271, 0.0275, 0.0251, 0.0198, 0.0142, 0.0088};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_11","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.034"
  cloud->annotation().addItem("Legend","x?F!=0.034");
  for(int index=0;index<6;index++) cloud->fill(xpm11[index],ypm11[index],1);
  aRegion.setParameter("plotter.pointsStyle[0].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[0].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[0].color",sxfColor[0]);
  aRegion.plot(*cloud);}

  double xpm12[] = {0., 0.5};
  double ypm12[] = {0.0112, 0.0047};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_12","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.15"
  cloud->annotation().addItem("Legend","x?F!=0.15");
  for(int index=0;index<2;index++) cloud->fill(xpm12[index],ypm12[index],1);
  aRegion.setParameter("plotter.pointsStyle[1].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[1].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[1].color",sxfColor[1]);
  aRegion.plot(*cloud);}

  double xpm13[] = {0., 0.5};
  double ypm13[] = {0.0017, 0.00085};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_13","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.50"
  cloud->annotation().addItem("Legend","x?F!=0.50");
  for(int index=0;index<2;index++) cloud->fill(xpm13[index],ypm13[index],1);
  aRegion.setParameter("plotter.pointsStyle[2].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[2].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[2].color",sxfColor[2]);
  aRegion.plot(*cloud);}

  double xpm14[] = {0.5};
  double ypm14[] = {0.000152};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_14","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.75"
  cloud->annotation().addItem("Legend","x?F!=0.75");
  for(int index=0;index<1;index++) cloud->fill(xpm14[index],ypm14[index],1);
  aRegion.plot(*cloud);}
  aRegion.setParameter("plotter.pointsStyle[3].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[3].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[3].color",sxfColor[3]);

  return true;
}

//////////////////////////////////////////////////////////////////////////////
bool plot_K_minus(
 AIDA::IPlotterRegion& aRegion
,const std::vector<AIDA::IDataPointSet*>& aDPS
,AIDA::IHistogramFactory& aHistogramFactory 
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  set_region_style(aRegion);

  aRegion.setParameter("plotter.yAxisAutomated","FALSE");
  aRegion.setParameter("plotter.yAxisMinimum","1.e-6");
  aRegion.setParameter("plotter.yAxisMaximum","1.e-1");
  //FIXME : handle "#K^{-}"
  aRegion.setTitle("K^-!");

  //  Color_t xfColor[] = {kGreen,kBlue,kRed,kBlack};
  const char* sxfColor[] = { 
    "0 1 0", //green
    "0 0 1", //blue
    "1 0 0", //red
    "0 0 0"  "black"
  };

  int ixf;
  for (ixf=0; ixf<numberOfXFValues; ++ixf) { //loop on xF ranges
    std::string sbin = "plotter.binsStyle[";
    sbin += Lib::smanip::tostring(ixf);
    sbin += "].";
    aRegion.setParameter(sbin+"visible","FALSE");
    //aRegion.setParameter(sbin+"modeling","wire_boxes");
    //aRegion.setParameter(sbin+"modeling","top_lines");
    //aRegion.setParameter(sbin+"color",sxfColor[ixf]);

    std::string serror = "plotter.errorsStyle[";
    serror += Lib::smanip::tostring(ixf);
    serror += "].";
    aRegion.setParameter(serror+"visible","TRUE");
    aRegion.setParameter(serror+"color",sxfColor[ixf]);
    aRegion.setParameter(serror+"lineWidth","2");
    aRegion.plot(*aDPS[ixf]);
  }//loop on xf values
  
  //exp data from Bonesini et al

  double xpm21[] = {0., 0.075,  0.225};
  double ypm21[] = {0.0214, 0.020,  0.0154};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_21","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.034"
  cloud->annotation().addItem("Legend","x?F!=0.034");
  for(int index=0;index<3;index++) cloud->fill(xpm21[index],ypm21[index],1);
  aRegion.setParameter("plotter.pointsStyle[0].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[0].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[0].color",sxfColor[0]);
  aRegion.plot(*cloud);}

  double xpm22[] = {0, 0.5};
  double ypm22[] = {0.00561, 0.00226};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_22","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.15"
  cloud->annotation().addItem("Legend","x?F!=0.15");
  for(int index=0;index<2;index++) cloud->fill(xpm22[index],ypm22[index],1);
  aRegion.setParameter("plotter.pointsStyle[1].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[1].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[1].color",sxfColor[1]);
  aRegion.plot(*cloud);}

  double xpm23[] = {0., 0.5};
  double ypm23[] = {0.000226,  0.000088};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_23","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.50"
  cloud->annotation().addItem("Legend","x?F!=0.50");
  for(int index=0;index<2;index++) cloud->fill(xpm23[index],ypm23[index],1);
  aRegion.setParameter("plotter.pointsStyle[2].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[2].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[2].color",sxfColor[2]);
  aRegion.plot(*cloud);}

  double xpm24[] = {0., 0.5};
  double ypm24[] = {2.82e-6, 1.36e-6};
 {AIDA::ICloud2D* cloud = aHistogramFactory.createCloud2D("cloud_24","");
  if(!cloud) return false;
  //FIXME : handle "x_{F}=0.75"
  cloud->annotation().addItem("Legend","x?F!=0.75");
  for(int index=0;index<2;index++) cloud->fill(xpm24[index],ypm24[index],1);
  aRegion.setParameter("plotter.pointsStyle[3].markerStyle","CIRCLE_FILLED");
  aRegion.setParameter("plotter.pointsStyle[3].markerSize","9");
  aRegion.setParameter("plotter.pointsStyle[3].color",sxfColor[3]);
  aRegion.plot(*cloud);}

  return true;
}

//////////////////////////////////////////////////////////////////////////////
void set_region_style(
 AIDA::IPlotterRegion& aRegion
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //FIXME : have to do the below with AIDA styles.

  // ROOT is in NDC, then we take a plotter with width = height = 1

  // PAW :
  //float XSIZ = 20.0F;
  //float XLAB = 1.4F;
  //float YSIZ = 20.0F;
  //float YLAB = 1.4F;

  float XLAB = 1.4F/20.0F; //0.07 //x distance of y title to data frame.
  float YLAB = 0.8F/20.0F; //0.04 //y distance of x title to data frame.
  YLAB = 0.05F; //FIXME : cooking.

  // To have a good matching with ROOT for text size :
  //double majic = 0.014/0.027;
  double majic = 0.6;
  std::string s = Lib::smanip::tostring(majic);
  //FIXME : have methods for the below :
  aRegion.style().setParameter("textScale",s);
  aRegion.style().xAxisStyle().setParameter("magStyle.scale",s);
  aRegion.style().yAxisStyle().setParameter("magStyle.scale",s);
  aRegion.style().xAxisStyle().tickLabelStyle().setParameter("scale",s);
  aRegion.style().xAxisStyle().labelStyle().setParameter("scale",s);
  aRegion.style().yAxisStyle().tickLabelStyle().setParameter("scale",s);
  aRegion.style().yAxisStyle().labelStyle().setParameter("scale",s);

  // ROOT def margins 0.1. SoPlotter def 0.1.
  aRegion.layout().setParameter("rightMargin",0.1);
  aRegion.layout().setParameter("topMargin",0.1);
  //gStyle->SetPadBottomMargin(0.15);
  aRegion.layout().setParameter("leftMargin",0.15);
  //gStyle->SetPadLeftMargin(0.15);
  aRegion.layout().setParameter("bottomMargin",0.15);

  aRegion.style().setParameter("superposeBins","TRUE");
  aRegion.setParameter("plotter.wallStyle.visible","FALSE");
  aRegion.setParameter("plotter.gridStyle.visible","FALSE");

  //std::string font = "Hershey"; //PAW font.
  //std::string font = "TTF/arialbd"; //Default ROOT font 62.
  //std::string font = "TTF/couri";
  std::string font = "TTF/times"; //= ROOT 132 ?
  //std::string smoothing = "FALSE";
  std::string smoothing = "TRUE";

  // X axis :
  aRegion.style().xAxisStyle().tickLabelStyle().setFont(font);
  aRegion.style().xAxisStyle().tickLabelStyle().setParameter("smoothing",smoothing);
  aRegion.style().xAxisStyle().labelStyle().setFont(font);
  aRegion.style().xAxisStyle().labelStyle().setParameter("smoothing",smoothing);
  aRegion.setParameter("plotter.xAxis.magStyle.fontName",font);
  aRegion.setParameter("plotter.xAxis.magStyle.smoothing",smoothing);
  aRegion.style().xAxisStyle().lineStyle().setThickness(2);
  aRegion.setParameter("plotter.xAxis.ticksStyle.width","2");
  aRegion.style().setParameter("topAxisVisible","TRUE");
  //FIXME : handle : "p_{T} (GeV)" ?
  aRegion.style().xAxisStyle().setLabel("p?T! (GeV)");
  // Set hplot tick modeling :
  aRegion.style().xAxisStyle().setParameter("modeling","hplot");
  aRegion.style().xAxisStyle().setParameter("divisions","505");

  // Y axis :
  aRegion.style().setParameter("yAxisLogScale","TRUE");
  aRegion.style().yAxisStyle().tickLabelStyle().setFont(font);
  aRegion.style().yAxisStyle().tickLabelStyle().setParameter("smoothing",smoothing);
  aRegion.style().yAxisStyle().labelStyle().setFont(font);
  aRegion.style().yAxisStyle().labelStyle().setParameter("smoothing",smoothing);
  aRegion.setParameter("plotter.yAxis.magStyle.fontName",font);
  aRegion.setParameter("plotter.yAxis.magStyle.smoothing",smoothing);
  aRegion.style().yAxisStyle().lineStyle().setThickness(2);
  aRegion.setParameter("plotter.yAxis.ticksStyle.width","2");

  //gStyle->SetTitleSize(0.06,"XYZ"); //ROOT def 0.04. SoPlotter : 0.014
  aRegion.setParameter("plotter.xAxis.titleHeight","0.06");
  aRegion.setParameter("plotter.yAxis.titleHeight","0.06");
  //gStyle->SetLabelOffset(0.01,"Y"); //ROOT def 0.005. SoPlotter def 0.02
  aRegion.setParameter("plotter.yAxis.labelToAxis","0.01");
  //gStyle->SetTitleOffset(1.1,"Y");
  aRegion.setParameter("plotter.yAxis.titleToAxis",
   Lib::smanip::tostring(1.1*XLAB));
  aRegion.setParameter("plotter.xAxis.titleToAxis",
   Lib::smanip::tostring(1.1*YLAB));
  //gStyle->SetLabelSize(0.05,"XYZ"); //ROOT def 0.04. SoPlotter def 0.014.
  aRegion.setParameter("plotter.yAxis.labelHeight","0.05");
  aRegion.setParameter("plotter.xAxis.labelHeight","0.05");

  aRegion.style().setParameter("rightAxisVisible","TRUE");
  //FIXME : handle "E d^{3}#sigma/dp^{3} (b/GeV^{2})" ?
  aRegion.style().yAxisStyle().setLabel("E d^3![s]/dp^3! (b/GeV^2!)");
  // Set hplot tick modeling :
  aRegion.style().yAxisStyle().setParameter("modeling","hplot");
  aRegion.style().yAxisStyle().setParameter("divisions","505");

  // title :
  aRegion.style().setParameter("titleHeight","0.04");
  aRegion.style().setParameter("titleToAxis","0.02");
  aRegion.style().titleStyle().textStyle().setFont(font);
  aRegion.style().titleStyle().textStyle().setParameter("smoothing",smoothing);
  aRegion.style().titleStyle().textStyle().setColor("0 0 1");

  // legend box :
  aRegion.setParameter("legendRegionVisible","TRUE");
  aRegion.setParameter("legendRegion.viewportRegion.backgroundColor","1 1 0");
  aRegion.setParameter("legendRegionSize","0.15 0.16");
  aRegion.setParameter("legendRegionOrigin","0.1 0.1");
  aRegion.setParameter("legendRegion.horizontalMargin","2");
  aRegion.setParameter("legendRegion.verticalMargin","6");

  // Frame :
  aRegion.setParameter("plotter.innerFrameStyle.lineWidth","2");
}
