#include <HEPVis/Gui/SoGui.h>

#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>

#include <HEPVis/SbArguments.h>
#include <HEPVis/SbMath.h>
#include <HEPVis/SbColors.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoTextHershey.h>
#include <HEPVis/nodes/SoTextTTF.h>
#include <HEPVis/nodes/SoGrid.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/nodes/SoLineStyle.h>
#include <HEPVis/nodekits/SoAxis.h>
#include <HEPVis/nodekits/SoAxes3D.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

#include <stdlib.h>

static void keystroke(void*,SoEventCallback*);
static void sensorCB(void*,SoSensor*);

static SoSeparator* sMeasuredScene = 0;
static SoScale* sMeasuredScale = 0;

static float sTrans = 0;

int main(int aArgc, char** aArgv) {

  SbArguments args(aArgc,aArgv);

  // Set up the main window
  SoGuiWidget topWidget=SoGui::init(aArgv[0]);
  if (!topWidget) exit(1);

  // initialize HEPVis used classes :
  SoTextHershey::initClass();
  SoTextTTF::initClass();
  SoTextStyle::initClass();
  SoLineStyle::initClass();
  SoAxis::initClass();
  SoAxes3D::initClass();
  SoGrid::initClass();
  SoAlternateRepAction::initClass();

  SoSeparator* root=new SoSeparator;
  root->ref();

  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback
    (SoKeyboardEvent::getClassTypeId(),keystroke,0);
  root->addChild(eventCallback);                                  

  SoLightModel* lightModel = new SoLightModel;
  //lightModel->model.setValue(SoLightModel::BASE_COLOR);
  root->addChild(lightModel);

  SoAxes3D* axes = new SoAxes3D;
  axes->getXAxis()->getLabelsStyle()->color.setValue(SbColor_yellow);
  axes->getYAxis()->getLabelsStyle()->color.setValue(SbColor_yellow);
  axes->getZAxis()->getLabelsStyle()->color.setValue(SbColor_yellow);
  
  axes->gridVisible.setValue(TRUE);
  axes->wallVisible.setValue(TRUE);
  axes->wallColor.setValue(SbColor(0.5f,0.5f,0.5f));

  axes->getXAxis()->getLineStyle()->pattern.setValue(SbLinePattern_dashed);
  axes->getYAxis()->getLineStyle()->pattern.setValue(SbLinePattern_dotted);
  axes->getZAxis()->getLineStyle()->pattern.setValue(SbLinePattern_dash_dotted);

  axes->getXAxis()->minimumValue.setValue(-13);
  axes->getXAxis()->maximumValue.setValue(20);
  axes->getXAxis()->title.setValue("AxisX3");

  axes->getYAxis()->minimumValue.setValue(0);
  axes->getYAxis()->maximumValue.setValue(5.5f);
  axes->getYAxis()->title.setValue("AxisY3");

  axes->getZAxis()->minimumValue.setValue(0);
  axes->getZAxis()->maximumValue.setValue(1);
  axes->getZAxis()->title.setValue("AxisZ3");
  root->addChild(axes);  

  // To avoid looping, sMeasuredScale must be BEFORE sMeasuredScene 
  // (the sensorCB will touch sMeasuredScale).
  sMeasuredScale = new SoScale;
  root->addChild(sMeasuredScale);

  sMeasuredScene = new SoSeparator;
  root->addChild(sMeasuredScene);

  SoNodeSensor* sensor = new SoNodeSensor(sensorCB,axes->getZAxis());
  //sensor->setPriority(0);
  //sensor->setTriggerPathFlag(TRUE);
  sensor->attach(sMeasuredScene);

  SoSeparator* sepCube = new SoSeparator;
  sMeasuredScene->addChild(sepCube);

  // Put base of the cube in the XY plan :
  SoTranslation* translation = new SoTranslation;
  translation->translation.setValue(SbVec3f(0.5F,0.5F,0.5F));
  sepCube->addChild(translation);

  SoCube* cube = new SoCube();
  cube->width = 0.1f;
  cube->height = 0.1f;
  cube->depth = 1.0f;
  sepCube->addChild(cube);

  if (args.isAnArgument("-write")) {
    SbBool altRep = args.isAnArgument("-altrep")?TRUE:FALSE;
    SbString file;
    args.find("-file",file);
    if(file=="") file = "out.iv";
    SoTools::write(*root,file,"",FALSE,altRep);
  }

  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(topWidget);
  examinerViewer->setSceneGraph(root);
  examinerViewer->viewAll();
  examinerViewer->show();

  SoGui::show(topWidget);

  SoDebugError::postInfo("Axes3D",
"\n\
 Press any key in picking mode to add objects in the scene.\n\
 The Z axis will rescale acoording the overall height of objects.");

  SoGui::mainLoop();

  root->unref();
  delete examinerViewer;

#ifdef SoFreePackage //Cleanup
  SoGui::clear();  
#endif

  return 0; 
}
void keystroke(void*,SoEventCallback* eventCB){
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if(SO_KEY_PRESS_EVENT(event,ANY)) {
    //SoDebugError::postInfo("Axes3D","key pressed...");

    SoSeparator* sepCylinder = new SoSeparator;
    sMeasuredScene->addChild(sepCylinder);

    float x = 0.1f + sTrans;
    float y = 0.1f + sTrans;
    float h = 2 + 2 * sTrans;
    sTrans += 0.1f;

    // Put base of the cube in the XY plan :
    SoTranslation* translation = new SoTranslation;
    translation->translation.setValue(SbVec3f(x,y,h/2));
    sepCylinder->addChild(translation);

    SoRotationXYZ* rotation = new SoRotationXYZ;
    rotation->angle = FM_PI_2;
    rotation->axis = SoRotationXYZ::X;
    sepCylinder->addChild(rotation);
    
    SoCylinder* cylinder = new SoCylinder();
    cylinder->radius = 0.05f;
    cylinder->height = h;
    sepCylinder->addChild(cylinder);

    eventCB->setHandled();
  }
}
static void sensorCB(void* aData,SoSensor*) {
  //SoDebugError::postInfo("Axes3D","sensorCB...");
  SoAxis* axisZ = (SoAxis*)aData;

  // Get BB of measuredScene :
  SbViewportRegion vp;
  SoGetBoundingBoxAction getBB(vp);
  getBB.apply(sMeasuredScene);
  SbBox3f bb = getBB.getBoundingBox();

  axisZ->minimumValue.setValue(bb.getMin()[2]);
  axisZ->maximumValue.setValue(bb.getMax()[2]);
 
  float bbHeight = bb.getMax()[2] - bb.getMin()[2]; 
  sMeasuredScale->scaleFactor.setValue(SbVec3f(1,1,1/bbHeight));
 
}
