#include <math.h>
#include <Inventor/SoInteraction.h>
#include <Inventor/SoInput.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodekits/SoAppearanceKit.h>
#include "HEPVis/nodes/SoHelicalTrack.h"
#include "HEPVis/nodes/SoTubs.h"
#include "HEPVis/nodes/SoBox.h"
#include "HEPVis/nodes/SoTrd.h"
#include "HEPVis/nodes/SoCons.h"
#include "HEPVis/nodes/SoTrap.h"
#include "HEPVis/nodes/SoSiStrips.h"
#include "HEPVis/nodekits/SoSiStripDetKit.h"
#include "HEPVis/nodekits/SoDblSiStripDetKit.h"
#include "HEPVis/nodekits/SoStereoSiStripDetKit.h"
#include "HEPVis/nodekits/SoDetectorTreeKit.h"
#include "HEPVis/nodes/SoEllipsoid.h"
#include "HEPVis/nodes/SoEllipticalPrism.h"
#include "HEPVis/nodes/SoReconTrack.h"
#include "HEPVis/nodes/SoVtxReconTrack.h"
#include "HEPVis/nodes/SoEllipsoidSegment.h"

#include <iostream>

// All HEPVis nodes have a generateAlternateRep method
// This template generates code that invokes that function
template <class HEPVisNode> 
void generateAlternateRep (SoGroup *parent, HEPVisNode * & node) {
  node->generateAlternateRep((SoAction*)0);
  SoNode *newNode= node->alternateRep.getValue()->copy();
  parent->replaceChild(node,newNode);
}

#define CONVERTHEPVIS(HEPVisType)                               \
  if (child->isOfType(HEPVisType::getClassTypeId())) {          \
    HEPVisType *hepvisInstance = (HEPVisType *) child;          \
    generateAlternateRep(parent,hepvisInstance);                \
    continue;                                                   \
  }
 
// Descend through the scene database and convert hepvis
// nodes to standard inventor.
void descendAndConvert(SoGroup * parent ) {
  for (int i=0;i<parent->getNumChildren();i++) {
    SoNode *child = parent->getChild(i);
    //
    // std::cout << child->getTypeId().getName().getString() << std::endl;
    //
    // If the node is a group node, decend its children
    //
    if (child->isOfType(SoGroup::getClassTypeId())) {
      descendAndConvert((SoGroup *) child);
    }
    //
    // If the node is a detector tree kit, create a replacement
    // separator.
    //
    else if (child->isOfType(SoDetectorTreeKit::getClassTypeId())) {
      SoDetectorTreeKit *detectorTreeKit=(SoDetectorTreeKit *) child;
      SoSeparator *replacementSeparator=new SoSeparator;
      //
      // Get the appearanceKit and the transform out of this guy
      //
      SoAppearanceKit * apKit= 
        (SoAppearanceKit *) detectorTreeKit->getPart("appearance",FALSE);
      SoTransform     * trans= 
        (SoTransform *)     detectorTreeKit->getPart("transform",FALSE);
      //
      // If these two objects, exist, add them to the replacement separator
      //
      if (apKit) replacementSeparator->addChild(apKit);
      if (trans) replacementSeparator->addChild(trans);
      //
      // Now decend and convert the children of the active separator
      //
      SoSeparator *activeSeparator=detectorTreeKit->getFullSeparator();
      if (detectorTreeKit->getPreview()) 
        activeSeparator=detectorTreeKit->getPreviewSeparator();
      descendAndConvert(activeSeparator);
      replacementSeparator->addChild(activeSeparator);
      //
      // Now finally swap the detector tree kit for its replacement separator
      //
      parent->replaceChild(detectorTreeKit,replacementSeparator);
    }
    else {
        CONVERTHEPVIS(SoHelicalTrack)
	CONVERTHEPVIS(SoTubs)
	CONVERTHEPVIS(SoTrd)
	CONVERTHEPVIS(SoBox)
	CONVERTHEPVIS(SoTrap)
	CONVERTHEPVIS(SoCons)
	CONVERTHEPVIS(SoSiStrips)
	CONVERTHEPVIS(SoReconTrack)
	CONVERTHEPVIS(SoVtxReconTrack)
	CONVERTHEPVIS(SoEllipsoid)
	CONVERTHEPVIS(SoEllipsoidSegment)
	CONVERTHEPVIS(SoEllipticalPrism)
    }
  }
}

// Call the initClass method on all HEPVis Extender nodes.
void initAllHEPVisClasses() {
  SoHelicalTrack::initClass();
  SoTubs::initClass();
  SoBox::initClass();
  SoTrd::initClass();
  SoCons::initClass();
  SoTrap::initClass();
  SoSiStrips::initClass();
  SoSiStripDetKit::initClass();
  SoDblSiStripDetKit::initClass();
  SoStereoSiStripDetKit::initClass();
  SoDetectorTreeKit::initClass();
  SoEllipsoid::initClass();
  SoEllipticalPrism::initClass();
  SoReconTrack::initClass();
  SoVtxReconTrack::initClass();
  SoEllipsoidSegment::initClass();
}

int main(int argc, char *argv[]) {

  // Initialization
  SoInteraction::init();
  initAllHEPVisClasses();

  // Parse the command line
  if (argc!=2) {
    std::cerr << "Usage: hepvisToStdInventor InputFile" << std::endl;
    return 1;
  }

  // Try to open the input file
  SoInput in;
  if (! in.openFile(argv[1], TRUE)) {
    std::cerr << "Error opening input file " << argv[1] << std::endl;
    return 1;
  }

  //See if it is a valid Inventor file
  if (! in.isValidFile()) {
    std::cerr << "Input file " << argv[1] << " is not a valid Inventor file" << std::endl;
    return 1;
  }
  
  //Read the file in
  SoSeparator *root=SoDB::readAll(&in);
  if (!root) return 1;
  root->ref();

  //Decend through the scene, converting any hepvis nodes to 
  //Standard Open Inventor
  descendAndConvert(root);

  //Write the converted HEPVis file to standard output
  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  return 0;
}
