// this :
#include <OnX/Net/NetManager.h>

#include <Slash/Core/IFileGetter.h>
#include <Slash/Core/ISession.h>

#include <Lib/Cast.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/InternetSocket.h>
#include <Lib/Debug.h>
#include <Lib/Out.h>
#include <Lib/sout.h>

#include <OnX/Net/Protocol.h>

#include <algorithm>

namespace OnX {
namespace Net {

class FileGetter : public virtual Slash::Core::IFileGetter {
public: //Slash::Core::IFileGetter:
  virtual bool fetch(const std::string& aFileName,
                     const Slash::Core::IFileGetter::Arguments&,
                     std::string& aNewName) {
    //::printf("debug : FileGetter::fetch : \"%s\"\n",aFileName.c_str());
    // Ask to the client the file :
    if(!fNetManager.askFile(aFileName)) return false;
    std::string path,name,suffix;
    Lib::smanip::pathNameSuffix(aFileName,path,name,suffix);
    aNewName = name;
    fFiles.push_back(aNewName);
    return fNetManager.waitFile(aNewName);
  }  
  virtual bool exists(const std::string& aFileName,bool& aValue) {
    return fNetManager.askFileExists(aFileName,aValue);
  }
public:
  FileGetter(Slash::Net::INetManager& aNetManager):fNetManager(aNetManager) {}
  virtual ~FileGetter() {
    std::for_each(fFiles.begin(),fFiles.end(),Lib::fmanip::remove);
  }
private:
  Slash::Net::INetManager& fNetManager;
  std::vector<std::string> fFiles;
};

}}

OnX::NetManager::NetManager(
 Slash::Core::ISession& aSession
,Lib::InternetSocket& aSocket
,bool aIsClient
)
:fSession(aSession)
,fName("NetManager")
,fSocket(aSocket)
,fIsClient(aIsClient)
,fFileGetter(0)
{
  Lib::Debug::increment("OnX::NetManager");
  fFileGetter = new OnX::Net::FileGetter(*this);
}

OnX::NetManager::~NetManager(){
  delete fFileGetter;
  Lib::Debug::decrement("OnX::NetManager");
}

std::string OnX::NetManager::name() const { return fName;}

void* OnX::NetManager::cast(const std::string& aClass) const {
  if_Lib_SCast(Slash::Net::INetManager)
  else if_Lib_SCast(Slash::Core::IManager)
  else return 0;
}


bool OnX::NetManager::isClient() const { return fIsClient;}
Slash::Core::IFileGetter* OnX::NetManager::fileGetter() const { return fFileGetter;}

bool OnX::NetManager::fetchFile(
 const std::string& aFileName
,std::string& aNewName) {
  Slash::Core::IFileGetter::Arguments args;
  return fFileGetter->fetch(aFileName,args,aNewName);
}

bool OnX::NetManager::executeScript(
 const std::string& aInterp
,const std::string& aScript
,const Slash::Net::INetManager::Options& //FIXME : pass params
) {
  if(!fSocket.isConnected()) return false;
  if(!fSocket.send_string
    (ONX_PROTOCOL_EXEC+":"+aInterp+":"+aScript)) return false;
  return true;
}
bool OnX::NetManager::pushEnd(const std::string& aWhat,const std::vector<std::string>& aArgs) {
  fEnds.push_back(End(aWhat,aArgs));
  return true;
}
bool OnX::NetManager::flush_one_end(const End& aEnd) {
  const std::string& what = aEnd.first;
  const std::vector<std::string>& args = aEnd.second;
  if(!fSocket.send_string(ONX_PROTOCOL_END+":"+what)) return false;
  if(!fSocket.send_text(args)) {
    Lib::Out out(fSession.printer());
    out << "OnX::NetManager::flush_one_end : "
        << " can't send end script " << Lib::sout(what) << "."
        << Lib::endl;
    return false;
  }
  if(what=="graphics") {
    const std::string& file = args[2];
    if(!sendFile(file)) return false;
    Lib::fmanip::remove(file);
  }

  return true;
}
bool OnX::NetManager::flushEnds() {
  if(!fSocket.isConnected()) return false;
  for(Ends::const_iterator it = fEnds.begin();it!=fEnds.end();++it) {
    if(!flush_one_end(*it)) return false;
  }
  fEnds.clear();
  return true;
}
unsigned int OnX::NetManager::sizeEnds() {return fEnds.size();}
bool OnX::NetManager::waitEnd(std::string& aWhat,std::vector<std::string>& aArgs) {
  if(!fSocket.isConnected()) return false;

  std::string s;
  if(!fSocket.wait_string(s)) return false;
  std::string what;
  std::string cmd;

  if(!OnX_protocol(s,what,cmd)) {
    Lib::Out out(fSession.printer());
    out << "OnX::NetManager::waitEnd : "
        << " bad syntax in " << Lib::sout(s) 
        << Lib::endl;
    fSocket.disconnect();
    return false;
  }

  if(what==ONX_PROTOCOL_END) {
    aWhat = cmd;
    if(!fSocket.wait_text(aArgs)) return false;
    return true;
  } 

  return false;
}

bool OnX::NetManager::sendFile(const std::string& aFile) {
  if(!fSocket.isConnected()) return false;
  if(!fSocket.send_string(ONX_PROTOCOL_FILE+":"+aFile)) return false;
  if(!fSocket.send_file(aFile)) {
    Lib::Out out(fSession.printer());
    out << "OnX::NetManager::sendFile : "
        << " can't send file " << Lib::sout(aFile) << "."
        << Lib::endl;
    return false;
  }
  return true;
}
bool OnX::NetManager::askFile(const std::string& aFile) {
  if(!fSocket.isConnected()) return false;
  return fSocket.send_string(ONX_PROTOCOL_FILE+":"+aFile);
}
bool OnX::NetManager::waitFile(const std::string& aFile) {
  if(!fSocket.isConnected()) return false;

  std::string s;
  if(!fSocket.wait_string(s)) return false;
  std::string what;
  std::string cmd;

  if(!OnX_protocol(s,what,cmd)) {
    Lib::Out out(fSession.printer());
    out << "OnX::NetManager::waitFile : "
        << " bad syntax in " << Lib::sout(s) 
        << Lib::endl;
    fSocket.disconnect();
    return false;
  }

  if(what==ONX_PROTOCOL_FILE) {
    //cmd not used.
    return fSocket.wait_file(aFile);
  }

  return false;
}

bool OnX::NetManager::askFileExists(const std::string& aFile,bool& aValue) {
  if(!fSocket.isConnected()) {aValue = false;return false;}
  if(!fSocket.send_string(ONX_PROTOCOL_FILE_EXISTS+":"+aFile)) {
    aValue = false;
    return false;
  }
  std::string s;
  if(!fSocket.wait_string(s)) {aValue = false;return false;}
  if(!Lib::smanip::tobool(s,aValue)) {
    Lib::Out out(fSession.printer());
    out << "OnX::NetManager::askFileExists : "
        << " received value not a boolean " << Lib::sout(s) 
        << Lib::endl;
    fSocket.disconnect();
    return false;
  }
  return true;
}


bool OnX::NetManager::sendExit(int a_code) {
  if(!fSocket.isConnected()) return false;
  std::string s = Lib::smanip::tostring(a_code);
  return fSocket.send_string(ONX_PROTOCOL_EXIT+":"+s);
}

bool OnX::NetManager::sendString(const std::string& aString) {
  if(!fSocket.isConnected()) return false;
  return fSocket.send_string(aString);
}
bool OnX::NetManager::waitString(std::string& aString) {
  if(!fSocket.isConnected()) return false;
  return fSocket.wait_string(aString);
}
