// this :
#include <oroot/internal/DLL.h>

#include <oroot/internal/mem.h>

#include <Slash/Tools/smanip.h>

#include <cstdlib>

#ifdef WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif

namespace oroot {

DLL::DLL(const std::string& aName,std::ostream& a_out,int aVerboseLevel,bool aQuiet)
:m_module(0)
,m_out(a_out)
,m_name(aName)
,m_verb(aVerboseLevel)
{
  // aName is the full name (path + libname).
  mem_inc("oroot::DLL");
  if(!open()) {
    if(!aQuiet) {
      m_out << "oroot::DLL::DLL :" 
            << " can't load " << m_name << " library." << std::endl;
      m_out << "oroot::DLL::DLL :" 
            << " error  : " << m_error 
            << std::endl;
    }
  }
}

DLL::~DLL(){
  if (m_module) {
    if(m_verb) {
      m_out << "oroot::DLL::~DLL :" 
            << " close " << Slash::sout(m_name) << "..." << std::endl;
    }
#ifdef WIN32
    ::FreeLibrary((HMODULE)m_module);
#else
    if(::dlclose(m_module)) {
      m_out << "oroot::DLL::~DLL :" 
            << " error  : " << ::dlerror() << std::endl;
    }
#endif
    if(m_verb) {
      m_out << "oroot::DLL::~DLL :" 
            << " close " << Slash::sout(m_name) << " done." << std::endl;
    }
  }
  m_module = 0;
  mem_dec("oroot::DLL");
}

//private:
DLL::DLL(const DLL& a_from):m_out(a_from.m_out){}
DLL& DLL::operator=(const DLL&){return *this;}

bool DLL::open() {
  m_error = "";
  std::string dll = m_name;  
  if(m_verb) {
    m_out << "oroot::DLL::DLL :" 
          << " open " 
          << Slash::sout(dll) << "..." << std::endl;
  }
#ifdef WIN32
  m_module = ::LoadLibrary(dll.c_str());
#else
 {std::string path,name,suffix;
  Slash::path_name_suffix(m_name,path,name,suffix);
  if(path.empty()) { //Relative path
    std::string suffix = Slash::suffix(dll);
    if(suffix.empty()) {
      // Only the name "Xxx" had been given, we have to
#if defined(__CYGWIN__) && defined(__GNUC__)
      // build the name "Xxx.dll" :      
      dll = m_name + ".dll";
#elif defined(__APPLE__)
      dll = m_name + ".bundle";
#else
      // build the name "libXxx.so" :      
      dll = "lib" + m_name + ".so";
#endif
    }
  }}
  int flags = RTLD_NOW | RTLD_GLOBAL;
  m_module = ::dlopen(dll.c_str(),flags);
  const char* cerror;
  if ((cerror=dlerror())!=NULL) m_error = cerror;
#endif
  if(!m_module) return false;
  if(m_verb) {
    m_out << "oroot::DLL::DLL :" 
          << " " << Slash::sout(dll) << " opened." << std::endl;
  }
  m_DLL = dll;  
  return true;
}

bool DLL::is_opened() const {return m_module?true:false;}

const std::string& DLL::file_name() const{return m_name;}

DLL::procedure DLL::find(const std::string& aName,bool aQuiet) const {
  if(!m_module) return 0;
  std::string name = aName;
  procedure p = 0;
  if(m_verb) {
    m_out << "oroot::DLL::find :" 
          << " search symbol " 
          << Slash::sout(name) << "..." << std::endl;
  }
#ifdef WIN32
  p = (procedure)::GetProcAddress((HMODULE)m_module,name.c_str());
  if (!p) {
    name = "_" + aName;
    p = (procedure)::GetProcAddress((HMODULE)m_module,name.c_str());
    if (!p) {
      if(!aQuiet) {
        m_out << "Lib : Can't find symbol " << Slash::sout(aName) << "." 
              << std::endl;
      }
    }
  }
#else
  p = (procedure)::dlsym(m_module,(char*)name.c_str());
  const char* cerror;
  if ((cerror=dlerror())!=NULL) {
    std::string serror = cerror;
    name = "_" + aName;
    p = (procedure)::dlsym(m_module,(char*)name.c_str());
    if ((cerror=dlerror())!=NULL) {
      if(!aQuiet) {
        m_out << "oroot::DLL::find :" 
              << " problem when searching " << Slash::sout(aName)
              << " in " << Slash::sout(m_DLL) << "."
              << " dlerror  : " << serror << std::endl;
      } 
    } 
  }
#endif
  if(p && m_verb) {
    m_out << "oroot::DLL::find :" 
          << " symbol " << Slash::sout(name) << " found." << std::endl;
  }
  return p;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool DLL::is_valid(const std::string& a_name,std::ostream& a_out) {
  DLL* dll = new DLL(a_name,a_out,0,true);
  bool status = dll->is_opened();
  delete dll;
  return status;
}


}
