#ifndef steve_HDF5_T_h
#define steve_HDF5_T_h

#include <string>
#include <vector>

#include <hdf5.h>

namespace steve {

namespace HDF5 {

template <class T>
bool write_array(
 hid_t a_location
,const std::string& a_name
,hid_t a_file_type
,hid_t a_mem_type
,unsigned int a_compression
,unsigned int a_size
,const T a_array[]
){
  if(a_compression>9) return false;
  unsigned int sz = a_size;
  if(!sz) return false; //NOTE : H5Screate_simple does not handle zero size.
  int dimn = 1;
  hsize_t* dims = new hsize_t[dimn];
  dims[0] = sz;
  hid_t dataspace = H5Screate_simple(dimn,dims,NULL); 
  delete [] dims;
  if(dataspace<0) return false;

  hid_t props = -1;
  if(a_compression) {
    props = H5Pcreate(H5P_DATASET_CREATE);
    if(props<0) {
      H5Sclose(dataspace);
      return false;
    }
    if(H5Pset_layout(props,H5D_CHUNKED)<0) { 
      H5Pclose(props);
      H5Sclose(dataspace);
      return false;
    }
    if(H5Pset_shuffle(props)<0) { //better compression.
      H5Pclose(props);
      H5Sclose(dataspace);
      return false;
    }
    hsize_t* cdims = new hsize_t[dimn];
    cdims[0] = sz;
    if(H5Pset_chunk(props,dimn,cdims)<0) {
      delete [] cdims;
      H5Pclose(props);
      H5Sclose(dataspace);
      return false;
    }
    delete [] cdims;
    if(H5Pset_deflate(props,a_compression)<0) {
      H5Pclose(props);
      H5Sclose(dataspace);
      return false;
    }
  } else {
    props = H5P_DEFAULT;
  }

  hid_t dataset = 
    H5Dcreate(a_location,a_name.c_str(),a_file_type,dataspace,props);
  if(dataset<0) {
    if(a_compression) H5Pclose(props);
    H5Sclose(dataspace);
    return false;
  }

  if(H5Dwrite(dataset,a_mem_type,H5S_ALL,H5S_ALL,H5P_DEFAULT,a_array)<0) {
    H5Dclose(dataset);                 
    if(a_compression) H5Pclose(props);
    H5Sclose(dataspace);
    return false;
  }

  H5Dclose(dataset);                 
  if(a_compression) H5Pclose(props);
  H5Sclose(dataspace);
  return true;
}

template <class T>
bool read_scalar(
 hid_t a_location
,const std::string& a_name
,hid_t a_mem_type
,T& a_data
){
  hid_t dataset = H5Dopen(a_location,a_name.c_str());
  if(dataset<0) return false;

  hid_t dataspace = H5Dget_space(dataset);
  if(dataspace<0) {
    H5Dclose(dataset);
    return false;
  }

  hid_t memspace = H5Screate(H5S_SCALAR); 
  if(memspace<0) {
    H5Sclose(dataspace);
    H5Dclose(dataset);
    return false;
  }

  if(H5Dread(dataset,a_mem_type,memspace,dataspace,H5P_DEFAULT,&a_data)<0) {
    H5Sclose(memspace);
    H5Sclose(dataspace);
    H5Dclose(dataset);
    return false;
  }


  H5Sclose(memspace);
  H5Sclose(dataspace);
  H5Dclose(dataset);

  return true;
}

template <class T>
bool read_array(
 hid_t a_location
,const std::string& a_name
,hid_t a_mem_type
,unsigned int& a_size
,T*& a_array
){
  hid_t dataset = H5Dopen(a_location,a_name.c_str());
  if(dataset<0) {
    a_size = 0;
    a_array = 0;
    return false; // data set not found.
  }

  hid_t dataspace = H5Dget_space(dataset);
  if(dataspace<0) {
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return false;
  }

  int dimn = H5Sget_simple_extent_ndims(dataspace);
  if(dimn<0) {
    H5Sclose(dataspace);
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return false;
  }
  if(dimn!=1) {
    H5Sclose(dataspace);
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return false;
  }
  //printf("debug : read dimn %d\n",dimn);

  hsize_t* dims = new hsize_t[dimn];
 {int rdimn = H5Sget_simple_extent_dims(dataspace,dims,NULL);
  if(rdimn<0) {
    delete [] dims;
    H5Sclose(dataspace);
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return false;
  }}

  unsigned int read_size = (unsigned int)dims[0];
  if(!read_size) {
    delete [] dims;
    H5Sclose(dataspace);
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return true; //It is ok.
  }

  hid_t memspace = H5Screate_simple(dimn,dims,NULL); 
  delete [] dims;
  if(memspace<0) {
    H5Sclose(dataspace);
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return false;
  }

  bool allocated = false;
  if(a_array) {
    // It is assumed that a_size contains the size of a_array.
    if(read_size>a_size) {
      H5Sclose(memspace);
      H5Sclose(dataspace);
      H5Dclose(dataset);
      a_size = 0;
      a_array = 0;
      return false;
    }
    a_size = read_size;
  } else {
    a_array = new T[read_size];
    a_size = read_size;
    allocated = true;
  }

  if(H5Dread(dataset,a_mem_type,memspace,dataspace,H5P_DEFAULT,a_array)<0) {
    if(allocated) delete [] a_array;
    H5Sclose(memspace);
    H5Sclose(dataspace);
    H5Dclose(dataset);
    a_size = 0;
    a_array = 0;
    return false;
  }

  H5Sclose(memspace);
  H5Sclose(dataspace);
  H5Dclose(dataset);

  return true;
}
  
} //hdf5

} //steve

#endif
