#ifndef Exp_Event_h
#define Exp_Event_h

#include <Exp/Exp.h>
#include <Exp/Params.h>

#include <Exp/instrument.h>

#include <steve/tools/blob.h>

namespace Exp {
  typedef unsigned int uint32;
  typedef int int32;
}

namespace Exp {class Analysis;}

#include <vector>
#include <string>

#include <HCL/Histogram.h>

namespace Exp {

class Time 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Time();
  Time(Params& a_params,
       Exp::uint32 a_indexInRun,
       Exp::uint32 a_indexInEvent,
       Exp::uint32 a_indexInTrack,
       Exp::uint32 a_indexInHit);
  virtual ~Time();
private:
  Time(const Time&);
  Time& operator=(const Time&);
public:
  Exp::int32 time() const;
  bool is_random() const;
private:
  Exp::int32 fTime;
  bool fRandom;
public:
  ClassDef(Time,1)
};

class Track;

class Hit 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Hit();
  Hit(Params& a_params,
      Track& a_track,
      Exp::uint32 a_indexInRun,
      Exp::uint32 a_indexInEvent,
      Exp::uint32 a_indexInTrack);
  virtual ~Hit();
private:
  Hit(const Hit&);
  Hit& operator=(const Hit&);
public:
  double x() const;
  double y() const;
  double z() const;
  std::vector<double> vec() const;
#ifdef EXP_REF
  Track* track() const;
#endif
  bool is_random() const;
private:
  double fX;
  double fY;
  double fZ;
  std::vector<double> fVec;
  bool fRandom;
public:
  unsigned int number_of_times() const;
  Exp::Time* time_at(unsigned int);
private:
  ARRAY_STAR_FIELD(fTimes,Exp::Time);
private:
#ifdef EXP_REF
  REF_FIELD(fTrack,Track);
#endif
public:
  ClassDef(Hit,1)
};

class Track 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Track();
  Track(Params& a_params,Exp::uint32 a_indexInRun,Exp::uint32 a_indexInEvent);
  virtual ~Track();
private:
  Track(const Track&);
  Track& operator=(const Track&);
public:
  double px() const;
  double py() const;
  double pz() const;
  bool is_random() const;
private:
  double fPx;
  double fPy;
  double fPz;
  bool fRandom;
public:
  unsigned int number_of_hits() const;
  Hit* hit_at(unsigned int);
private:
  ARRAY_STAR_FIELD(fHits,Hit);
public:
  ClassDef(Track,1)
};

class Cell 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Cell();
  Cell(Params& a_params,Exp::uint32 a_indexInRun,Exp::uint32 a_indexInEvent);
  virtual ~Cell();
  Cell(const Cell&);
  Cell& operator=(const Cell&);
#ifdef EXP_STRING
  std::string id() const;
  std::string id_vec() const;
private:
  std::string fID;
  std::string fID_Vec;
#endif
public:
  ClassDef(Cell,1)
};

// object to be streamed in a blob :
class Object 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Object();
  virtual ~Object();
private:
  Object(const Object&);
  Object& operator=(const Object&);
public:
  void initialize();
  steve::uchar get_uchar() const;
  char get_char() const;
  steve::uint16 get_uint16() const;
  steve::int16 get_int16() const;
  steve::uint32 get_uint32() const;
  steve::int32 get_int32() const;
  steve::uint64 get_uint64() const;
  steve::int64 get_int64() const;
  float get_float() const;
  double get_double() const;
  std::string get_string() const;
  bool is_equal(const Object&) const;
private:
  steve::uchar f_uchar;
  char f_char;
  steve::uint16 f_uint16;
  steve::int16 f_int16;
  steve::uint32 f_uint32;
  steve::int32 f_int32;
  steve::uint64 f_uint64;
  steve::int64 f_int64;
  float f_float;
  double f_double;
  std::string f_string;
public:
  ClassDef(Object,1)
};

class Event 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Event();
  Event(Params& a_params,
        Exp::uint32 a_indexInRun,Analysis* a_analysis = 0);
  virtual ~Event();
  unsigned int id() const;  
  float energy() const;  
  HCL::Histogram1D& histo();
  steve::tools::blob& blob();
  Object& object();
private:
  Event(const Event&);
  Event& operator=(const Event&);
private:
  Analysis* fAnalysis; //! for the pain.
  unsigned int fID;
  float fEnergy;
  HCL::Histogram1D fHisto; //! FIXME
  steve::tools::blob fBlob; //! FIXME
  Object fObject;
public:
  unsigned int number_of_tracks() const;
  Track* track_at(unsigned int);
  unsigned int number_of_cells() const;
  Cell* cell_at(unsigned int);
private:
  ARRAY_STAR_FIELD(fTracks,Track);
  ARRAY_FIELD(fCells,Cell);
#ifdef STEVE_USE_ROOT
public:
  bool store(steve::interface::store&,steve::interface::schema&) const;
  bool retrieve(steve::interface::store&,steve::interface::schema&,
                steve::uint32,const steve::path_t&);
#endif
public:
  ClassDef(Event,1)
};

class Header 
#ifdef STEVE_USE_ROOT
:public TObject
#endif
{
public:
  Header();
  virtual ~Header();
private:
  Header(const Header&);
  Header& operator=(const Header&);
public:
  unsigned int numberOfEvents() const;
  void setNumberOfEvents(unsigned int);
private:
  unsigned int fNumberOfEvents;
#ifdef STEVE_USE_ROOT
public:
  bool store(steve::interface::store&,steve::interface::schema&) const;
  bool retrieve(steve::interface::store&,steve::interface::schema&,
                steve::uint32,const steve::path_t&);
#endif
public:
  ClassDef(Header,1)
};

}

#endif
